# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from databricks_cli.unity_catalog.uc_service import UnityCatalogService
from databricks_cli.unity_catalog.utils import mc_pretty_format


class UnityCatalogApi(object):
    def __init__(self, api_client):
        self.client = UnityCatalogService(api_client)

    # Metastore APIs

    def create_metastore(self, name, storage_root, region):
        return self.client.create_metastore(name, storage_root, region)

    def list_metastores(self):
        return self.client.list_metastores()

    def get_metastore(self, metastore_id):
        return self.client.get_metastore(metastore_id)

    def update_metastore(self, metastore_id, metastore_spec):
        return self.client.update_metastore(metastore_id, metastore_spec)

    def delete_metastore(self, metastore_id, force):
        return self.client.delete_metastore(metastore_id, force)

    def get_metastore_summary(self):
        return self.client.get_metastore_summary()

    def create_metastore_assignment(self, workspace_id, metastore_id, default_catalog_name):
        return self.client.create_metastore_assignment(workspace_id, metastore_id,
                                                       default_catalog_name)

    def update_metastore_assignment(self, workspace_id, metastore_id, default_catalog_name):
        return self.client.update_metastore_assignment(workspace_id, metastore_id,
                                                       default_catalog_name)

    def delete_metastore_assignment(self, workspace_id, metastore_id):
        return self.client.delete_metastore_assignment(workspace_id, metastore_id)

    def get_current_metastore_assignment(self):
        return self.client.get_current_metastore_assignment()

    # External Location APIs

    def create_external_location(self, loc_spec, skip_validation):
        return self.client.create_external_location(loc_spec, skip_validation)

    def list_external_locations(self):
        return self.client.list_external_locations()

    def get_external_location(self, name):
        return self.client.get_external_location(name)

    def update_external_location(self, name, loc_spec, force, skip_validation):
        return self.client.update_external_location(name, loc_spec, force, skip_validation)

    def delete_external_location(self, name, force):
        return self.client.delete_external_location(name, force)

    def validate_external_location(self, validation_spec):
        return self.client.validate_external_location(validation_spec)

    # Data Access Configuration APIs

    def create_dac(self, metastore_id, dac_spec, skip_validation):
        return self.client.create_dac(metastore_id, dac_spec, skip_validation)

    def list_dacs(self, metastore_id):
        return self.client.list_dacs(metastore_id)

    def get_dac(self, metastore_id, dac_id):
        return self.client.get_dac(metastore_id, dac_id)

    def delete_dac(self, metastore_id, dac_id):
        return self.client.delete_dac(metastore_id, dac_id)

    # Storage Credentials

    def create_storage_credential(self, cred_spec, skip_validation):
        return self.client.create_storage_credential(cred_spec, skip_validation)

    def list_storage_credentials(self, name_pattern):
        return self.client.list_storage_credentials(name_pattern)

    def get_storage_credential(self, name):
        return self.client.get_storage_credential(name)

    def update_storage_credential(self, name, cred_spec, skip_validation):
        return self.client.update_storage_credential(name, cred_spec, skip_validation)

    def delete_storage_credential(self, name, force):
        return self.client.delete_storage_credential(name, force)

    # Catalog APIs

    def create_catalog(self, catalog_name, comment, provider, share):
        return self.client.create_catalog(catalog_name, comment, provider, share)

    def list_catalogs(self):
        return self.client.list_catalogs()

    def get_catalog(self, name):
        return self.client.get_catalog(name)

    def update_catalog(self, name, catalog_spec):
        return self.client.update_catalog(name, catalog_spec)

    def delete_catalog(self, catalog_name):
        return self.client.delete_catalog(catalog_name)

    # Schema APIs

    def create_schema(self, catalog_name, schema_name, comment):
        return self.client.create_schema(catalog_name, schema_name, comment)

    def list_schemas(self, catalog_name, name_pattern):
        return self.client.list_schemas(catalog_name, name_pattern)

    def get_schema(self, full_name):
        return self.client.get_schema(full_name)

    def update_schema(self, full_name, schema_spec):
        return self.client.update_schema(full_name, schema_spec)

    def delete_schema(self, schema_full_name):
        return self.client.delete_schema(schema_full_name)

    # Table APIs

    def create_table(self, table_spec):
        return self.client.create_table(table_spec)

    def list_tables(self, catalog_name, schema_name, name_pattern):
        return self.client.list_tables(catalog_name, schema_name, name_pattern)

    def list_table_summaries(self, catalog_name):
        return self.client.list_table_summaries(catalog_name)

    def get_table(self, full_name):
        return self.client.get_table(full_name)

    def update_table(self, full_name, table_spec):
        return self.client.update_table(full_name, table_spec)

    def delete_table(self, table_full_name):
        return self.client.delete_table(table_full_name)

    # Share APIs

    def create_share(self, name):
        return self.client.create_share(name)

    def list_shares(self):
        return self.client.list_shares()

    def get_share(self, name, include_shared_data):
        return self.client.get_share(name, include_shared_data)

    def update_share(self, name, share_spec):
        return self.client.update_share(name, share_spec)

    def delete_share(self, name):
        return self.client.delete_share(name)

    def list_share_permissions(self, name):
        return self.client.list_share_permissions(name)

    def update_share_permissions(self, name, perm_spec):
        return self.client.update_share_permissions(name, perm_spec)

    # Recipient APIs

    def create_recipient(self, name, comment, sharing_id,
                         allowed_ip_addresses, custom_properties):
        return self.client.create_recipient(name, comment, sharing_id,
                                            allowed_ip_addresses, custom_properties)

    def list_recipients(self):
        return self.client.list_recipients()

    def get_recipient(self, name):
        return self.client.get_recipient(name)

    def update_recipient(self, name, recipient_spec):
        return self.client.update_recipient(name, recipient_spec)

    def rotate_recipient_token(self, name, existing_token_expire_in_seconds):
        return self.client.rotate_recipient_token(name, existing_token_expire_in_seconds)

    def get_recipient_share_permissions(self, name):
        return self.client.get_recipient_share_permissions(name)

    def delete_recipient(self, name):
        return self.client.delete_recipient(name)

    # Provider APIs

    def create_provider(self, name, comment, recipient_profile):
        return self.client.create_provider(name, comment, recipient_profile)

    def list_providers(self):
        return self.client.list_providers()

    def get_provider(self, name):
        return self.client.get_provider(name)

    def update_provider(self, name, new_name=None, comment=None,
                        recipient_profile=None, provider_spec=None):
        # If spec is provided, override all legacy parameters.
        if provider_spec is not None:
            return self.client.update_provider(name, provider_spec)
        _data = {}
        if new_name is not None:
            _data['name'] = new_name
        if recipient_profile is not None:
            _data['recipient_profile_str'] = mc_pretty_format(recipient_profile)
        if comment is not None:
            _data['comment'] = comment
        return self.client.update_provider(name, _data)
        
    def delete_provider(self, name):
        return self.client.delete_provider(name)

    def list_provider_shares(self, name):
        return self.client.list_provider_shares(name)

    # Permissions APIs

    def get_permissions(self, sec_type, sec_name):
        return self.client.get_permissions(sec_type, sec_name)

    def get_effective_permissions(self, sec_type, sec_name):
        return self.client.get_effective_permissions(sec_type, sec_name)

    def update_permissions(self, sec_type, sec_name, diff_spec):
        return self.client.update_permissions(sec_type, sec_name, diff_spec)

    def replace_permissions(self, sec_type, sec_name, perm_spec):
        return self.client.replace_permissions(sec_type, sec_name, perm_spec)

    # Lineage APIs

    def list_lineages_by_table(self, table_name):
        return self.client.list_lineages_by_table(table_name)

    def list_lineages_by_column(self, table_name, column_name):
        return self.client.list_lineages_by_column(table_name, column_name)



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from databricks_cli.sdk import SecretService


class SecretApi(object):
    def __init__(self, api_client):
        self.client = SecretService(api_client)

    def create_scope(self, scope, initial_manage_principal, scope_backend_type,
                     backend_azure_keyvault):
        return self.client.create_scope(scope, initial_manage_principal,
                                        scope_backend_type, backend_azure_keyvault)

    def delete_scope(self, scope):
        return self.client.delete_scope(scope)

    def list_scopes(self):
        return self.client.list_scopes()

    def put_secret(self, scope, key, string_value, bytes_value):
        return self.client.put_secret(scope, key, string_value, bytes_value)

    def delete_secret(self, scope, key):
        return self.client.delete_secret(scope, key)

    def list_secrets(self, scope):
        return self.client.list_secrets(scope)

    def put_acl(self, scope, principal, permission):
        return self.client.put_acl(scope, principal, permission)

    def delete_acl(self, scope, principal):
        return self.client.delete_acl(scope, principal)

    def list_acls(self, scope):
        return self.client.list_acls(scope)

    def get_acl(self, scope, principal):
        return self.client.get_acl(scope, principal)



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import os
from os import path

import click
from click import ParamType

from databricks_cli.oauth.oauth import get_tokens

from databricks_cli.configure.config import profile_option, get_profile_from_context, debug_option
from databricks_cli.configure.provider import DatabricksConfig, update_and_persist_config, \
    ProfileConfigProvider
from databricks_cli.sdk.version import API_VERSION, API_VERSIONS
from databricks_cli.utils import CONTEXT_SETTINGS

PROMPT_HOST = 'Databricks Host (should begin with https://)'
PROMPT_USERNAME = 'Username'
PROMPT_PASSWORD = 'Password'  # NOQA
PROMPT_TOKEN = 'Token'  # NOQA
ENV_AAD_TOKEN = 'DATABRICKS_AAD_TOKEN'
DEFAULT_SCOPES = [
    'accounts',
    'clusters',
    'mlflow',
    'offline_access',
    'sql',
    'unity-catalog'
]


def _configure_cli_token_file(profile, token_file, host, insecure, jobs_api_version):
    if not path.exists(token_file):
        raise RuntimeError('Unable to read token from "{}"'.format(token_file))

    with io.open(token_file, encoding='utf-8') as f:
        token = f.readline().strip()

    config = ProfileConfigProvider(profile).get_config() or DatabricksConfig.empty()
    if not host:
        host = click.prompt(PROMPT_HOST, default=config.host, type=_DbfsHost())

    new_config = DatabricksConfig.from_token(host=host,
                                             token=token,
                                             refresh_token=None,
                                             insecure=insecure,
                                             jobs_api_version=jobs_api_version)
    update_and_persist_config(profile, new_config)


def _configure_cli_token(profile, insecure, host, jobs_api_version):
    config = ProfileConfigProvider(profile).get_config() or DatabricksConfig.empty()

    if not host:
        host = click.prompt(PROMPT_HOST, default=config.host, type=_DbfsHost())

    token = click.prompt(PROMPT_TOKEN, default=config.token, hide_input=True)
    new_config = DatabricksConfig.from_token(host=host,
                                             token=token,
                                             refresh_token=None,
                                             insecure=insecure,
                                             jobs_api_version=jobs_api_version)
    update_and_persist_config(profile, new_config)


def scope_format(user_input):
    user_inputs = user_input.split(',')
    choice = click.Choice(DEFAULT_SCOPES)
    return list(map(lambda x: choice(x.strip()), user_inputs))


def _configure_cli_oauth(profile, insecure, host, scope, jobs_api_version):
    config = ProfileConfigProvider(profile).get_config() or DatabricksConfig.empty()

    if not host:
        host = click.prompt(PROMPT_HOST, default=config.host, type=_DbfsHost())
    if not scope:
        scope = click.prompt(
            "Pick one or more comma-separated scopes from {scopes}".format(scopes=DEFAULT_SCOPES),
            value_proc=scope_format,
            type=click.Choice(DEFAULT_SCOPES), default=None)
    access_token, refresh_token = get_tokens(host, scope)
    new_config = DatabricksConfig.from_token(host=host,
                                             token=access_token,
                                             refresh_token=refresh_token,
                                             insecure=insecure,
                                             jobs_api_version=jobs_api_version)
    update_and_persist_config(profile, new_config)


def _configure_cli_aad_token(profile, insecure, host, jobs_api_version):
    config = ProfileConfigProvider(profile).get_config() or DatabricksConfig.empty()

    if ENV_AAD_TOKEN not in os.environ:
        click.echo('[ERROR] Set Environment Variable \'%s\' with your '
                   'AAD Token and run again.\n' % ENV_AAD_TOKEN)
        click.echo('Commands to run to get your AAD token:\n'
                   '\t az login\n'
                   '\t token_response=$(az account get-access-token '
                   '--resource 2ff814a6-3304-4ab8-85cb-cd0e6f879c1d)\n'
                   '\t export %s=$(jq .accessToken -r <<< "$token_response")\n' % ENV_AAD_TOKEN
                   )
        return

    if not host:
        host = click.prompt(PROMPT_HOST, default=config.host, type=_DbfsHost())

    aad_token = os.environ.get(ENV_AAD_TOKEN)
    new_config = DatabricksConfig.from_token(host=host,
                                             token=aad_token,
                                             refresh_token=None,
                                             insecure=insecure,
                                             jobs_api_version=jobs_api_version)
    update_and_persist_config(profile, new_config)


def _configure_cli_password(profile, insecure, host, jobs_api_version):
    config = ProfileConfigProvider(profile).get_config() or DatabricksConfig.empty()
    if config.password:
        default_password = '*' * len(config.password)
    else:
        default_password = None

    if not host:
        host = click.prompt(PROMPT_HOST, default=config.host, type=_DbfsHost())

    username = click.prompt(PROMPT_USERNAME, default=config.username)
    password = click.prompt(PROMPT_PASSWORD, default=default_password, hide_input=True,
                            confirmation_prompt=True)
    if password == default_password:
        password = config.password
    new_config = DatabricksConfig.from_password(host, username, password, insecure,
                                                jobs_api_version)
    update_and_persist_config(profile, new_config)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Configures host and authentication info for the CLI.')
@click.option('--oauth', '-o', 'oauth', show_default=True, is_flag=True, default=False,
              help='[Experimental] Use OAuth for authorization')
@click.option('--scope', '-s', 'scope', show_default=True, multiple=True,
              type=click.Choice(DEFAULT_SCOPES),
              help='[Experimental] Specify which OAuth scopes to request access for')
@click.option('--token', '-t', 'token', show_default=True, is_flag=True, default=False)
@click.option('--token-file', '-f', 'token_file', default=None,
              help='Instead of reading the token from stdin, ' +
                   'read the token from a file provided by a secret store.')
@click.option('--host', show_default=True, default=None,
              help='Host to connect to.')
@click.option('--aad-token', show_default=True, is_flag=True, default=False)
@click.option('--insecure', show_default=True, is_flag=True, default=None,
              help='DO NOT verify SSL Certificates')
@click.option('--jobs-api-version', show_default=True, default=API_VERSION,
              type=click.Choice(API_VERSIONS), help='API version to use for jobs.')
@debug_option
@profile_option
def configure_cli(token, aad_token, insecure, host, token_file, jobs_api_version, oauth, scope):
    """
    Configures host, authentication, and jobs-api version for the CLI.
    """
    profile = get_profile_from_context()
    insecure_str = str(insecure) if insecure is not None else None

    if token:
        _configure_cli_token(profile=profile, insecure=insecure_str, host=host,
                             jobs_api_version=jobs_api_version)
    elif token_file:
        _configure_cli_token_file(profile=profile, insecure=insecure_str, host=host,
                                  token_file=token_file, jobs_api_version=jobs_api_version)
    elif oauth:
        _configure_cli_oauth(profile=profile, insecure=insecure_str, host=host,
                             scope=scope, jobs_api_version=jobs_api_version)
    elif aad_token:
        _configure_cli_aad_token(profile=profile, insecure=insecure_str, host=host,
                                 jobs_api_version=jobs_api_version)
    else:
        _configure_cli_password(profile=profile, insecure=insecure_str, host=host,
                                jobs_api_version=jobs_api_version)


class _DbfsHost(ParamType):
    """
    Used to validate the configured host
    """

    def convert(self, value, param, ctx):
        if value.startswith('https://'):
            return value
        else:
            self.fail('The host does not start with https://')



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import click

from databricks_cli.click_types import JsonClickType
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.unity_catalog.api import UnityCatalogApi
from databricks_cli.unity_catalog.utils import del_none, hide, json_file_help, json_string_help, \
    mc_pretty_format
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, json_cli_base


def create_update_common_options(f):
    @click.option('--url', default=None,
                help='Path URL for the new external location')
    @click.option('--storage-credential-name', default=None,
                help='Name of storage credential to use with new external location')
    @click.option('--read-only/--no-read-only', is_flag=True, default=None,
                help='Whether the external location is read-only')
    @click.option('--comment', default=None,
                help='Free-form text description.')
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        f(*args, **kwargs)
    return wrapper


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Create External Location.')
@click.option('--name', default=None,
              help='Name of new external location')
@create_update_common_options
@click.option('--skip-validation', '-s', 'skip_val', is_flag=True, default=False,
              help='Skip the validation of location\'s storage credential before creation')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='POST', path='/external-locations'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='POST', path='/external-locations'))
@debug_option
@profile_option
# UC's createExternalLocation returns a 401 when the validation of the external location's
# storage credential fails; that translates to a misleading error when eat_exceptions is enabled:
#   Your authentication information may be incorrect. Please reconfigure with ``dbfs configure``
# Until that is fixed (should return a 400), show full error trace.
#@eat_exceptions
@provide_api_client
def create_location_cli(api_client, name, url, storage_credential_name,
                        read_only, comment, skip_val, json_file, json):
    """
    Create new external location.

    The public specification for the JSON request is in development.
    """
    if ((name is not None) or (url is not None) or (storage_credential_name is not None) or
        (read_only is not None) or (comment is not None)):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other creation flags are specified')
        data = {
            'name': name,
            'url': url,
            'credential_name': storage_credential_name,
            'read_only': read_only,
            'comment': comment
        }
        loc_json = UnityCatalogApi(api_client).create_external_location(data, skip_val)
        click.echo(mc_pretty_format(loc_json))
    elif (json is None) and (json_file is None):
        raise ValueError('Must provide name, url and storage-credential-name' +
                         ' or use JSON specification')
    else:
        json_cli_base(json_file, json,
                      lambda json:
                      UnityCatalogApi(api_client).create_external_location(json, skip_val),
                      encode_utf8=True)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List external locations.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_locations_cli(api_client, ):
    """
    List external locations.
    """
    locs_json = UnityCatalogApi(api_client).list_external_locations()
    click.echo(mc_pretty_format(locs_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Get an external location.')
@click.option('--name', required=True,
              help='Name of the external location to get.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_location_cli(api_client, name):
    """
    Get an external location.
    """
    loc_json = UnityCatalogApi(api_client).get_external_location(name)
    click.echo(mc_pretty_format(loc_json))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Update an external location.')
@click.option('--name', required=True,
              help='Name of the external location to update.')
@click.option('--new-name', default=None, help='New name of the external location.')
@create_update_common_options
@click.option('--owner', default=None,
              help='Owner of the external location.')
@click.option('--force', '-f', is_flag=True, default=False,
              help='Force update even if location has dependent tables/mounts')
@click.option('--skip-validation', '-s', 'skip_val', is_flag=True, default=False,
              help='Skip the validation of location\'s storage credential')
@click.option('--json-file', default=None, type=click.Path(),
              help=json_file_help(method='PATCH', path='/external-locations/{name}'))
@click.option('--json', default=None, type=JsonClickType(),
              help=json_string_help(method='PATCH', path='/external-locations/{name}'))
@debug_option
@profile_option
# See comment for create_location_cli
#@eat_exceptions
@provide_api_client
def update_location_cli(api_client, name, new_name, url, storage_credential_name, read_only,
                        comment, owner, force, skip_val, json_file, json):
    """
    Update an external location.

    The public specification for the JSON request is in development.
    """
    if ((new_name is not None) or (storage_credential_name is not None) or
        (read_only is not None) or (comment is not None)):
        if (json_file is not None) or (json is not None):
            raise ValueError('Cannot specify JSON if any other update flags are specified')
        data = {
            'name': new_name,
            'url': url,
            'credential_name': storage_credential_name,
            'read_only': read_only,
            'comment': comment,
            'owner': owner
        }
        loc_json = UnityCatalogApi(api_client).update_external_location(
            name, data, force, skip_val)
        click.echo(mc_pretty_format(loc_json))
    else:
        json_cli_base(json_file, json,
                      lambda json: UnityCatalogApi(api_client).update_external_location(name, json,
                                                                                        force,
                                                                                        skip_val),
                      encode_utf8=True)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Delete an external location.')
@click.option('--name', required=True,
              help='Name of the external location to delete.')
@click.option('--force', '-f', is_flag=True, default=False,
              help='Force deletion even if location has dependent tables/mounts')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_location_cli(api_client, name, force):
    """
    Delete an external location.
    """
    UnityCatalogApi(api_client).delete_external_location(name, force)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Validate a external location/credential pair')
@click.option('--name', default=None,
              help='Name of the external location to validate.')
@click.option('--url', default=None,
              help='A storage URL to validate.')
@click.option('--cred-name', default=None,
              help='Name of the storage credential to use for validation.')
@click.option('--cred-aws-iam-role', default=None,
              help='An aws role to validate')
@click.option('--cred-az-directory-id', default=None,
              help='An Azure Service Principal directory id to validate')
@click.option('--cred-az-application-id', default=None,
              help='An Azure Service Principal application id to validate')
@click.option('--cred-az-client-secret', default=None,
              help='An Azure Service Principal directory id to validate')
@click.option('--cred-az-mi-access-connector-id', default=None,
              help='An Azure Managed Identity access connector id to validate')
@click.option('--cred-az-mi-id', default=None,
              help='An Azure Managed Identity id to validate')
@click.option('--cred-gcp-sak-email', default=None,
              help='A GCP Service Account Key email to validate')
@click.option('--cred-gcp-sak-private-key-id', default=None,
              help='A GCP Service Account Key private key ID to validate')
@click.option('--cred-gcp-sak-private-key', default=None,
              help='A GCP Service Account Key private key to validate')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def validate_location_cli(api_client, name, url, cred_name, cred_aws_iam_role, cred_az_directory_id,
                          cred_az_application_id, cred_az_client_secret,
                          cred_az_mi_access_connector_id,
                          cred_az_mi_id, cred_gcp_sak_email, cred_gcp_sak_private_key_id,
                          cred_gcp_sak_private_key):
    """
    Validate an external location/credential combination.

    This call will attempt to read/list/write/delete with the given credentials and
    external location.

    One of name/url must be provided. If both are specified, the given credential
    name will be excluded from path overlap checks (used to validate a potential
    update of that credential).

    One of cred-name, or cloud provider specific credential parameters must be
    provided.
    """
    validation_spec = {
        'external_location_name': name,
        'url': url,
        'storage_credential_name': cred_name,
    }
    if cred_aws_iam_role is not None:
        validation_spec['aws_iam_role'] = {
            'role_arn': cred_aws_iam_role
        }

    if ((cred_az_directory_id is not None) or (cred_az_application_id is not None) or
        (cred_az_client_secret is not None)):
        validation_spec['azure_service_principal'] = {
            'directory_id': cred_az_directory_id,
            'application_id': cred_az_application_id,
            'client_secret': cred_az_client_secret
        }

    if (cred_az_mi_access_connector_id is not None) or (cred_az_mi_id is not None):
        validation_spec['azure_managed_identity'] = {
            'access_connector_id': cred_az_mi_access_connector_id,
            'managed_identity_id': cred_az_mi_id
        }

    if ((cred_gcp_sak_email is not None) or (cred_gcp_sak_private_key_id is not None) or
        (cred_gcp_sak_private_key is not None)):
        validation_spec['gcp_service_account_key'] = {
            'email': cred_gcp_sak_email,
            'private_key_id': cred_gcp_sak_private_key_id,
            'private_key': cred_gcp_sak_private_key
        }

    del_none(validation_spec)
    validation_json = UnityCatalogApi(api_client).validate_external_location(validation_spec)
    click.echo(mc_pretty_format(validation_json))


@click.group()
def external_locations_group():  # pragma: no cover
    pass


def register_ext_loc_commands(cmd_group):
    # Register deprecated "verb-noun" commands for backward compatibility.
    cmd_group.add_command(hide(create_location_cli), name='create-external-location')
    cmd_group.add_command(hide(list_locations_cli), name='list-external-locations')
    cmd_group.add_command(hide(get_location_cli), name='get-external-location')
    cmd_group.add_command(hide(update_location_cli), name='update-external-location')
    cmd_group.add_command(hide(delete_location_cli), name='delete-external-location')
    cmd_group.add_command(hide(validate_location_cli), name='validate-external-location')

    # Register command group.
    external_locations_group.add_command(create_location_cli, name='create')
    external_locations_group.add_command(list_locations_cli, name='list')
    external_locations_group.add_command(get_location_cli, name='get')
    external_locations_group.add_command(update_location_cli, name='update')
    external_locations_group.add_command(delete_location_cli, name='delete')
    external_locations_group.add_command(validate_location_cli, name='validate')
    cmd_group.add_command(external_locations_group, name='external-locations')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import click
from tabulate import tabulate

from databricks_cli.click_types import OutputClickType, SecretScopeClickType, SecretKeyClickType, \
    SecretPrincipalClickType
from databricks_cli.secrets.api import SecretApi
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, pretty_format, truncate_string, \
    error_and_quit
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.version import print_version_callback, version


SCOPE_HEADER = ('Scope', 'Backend', 'KeyVault URL')
SECRET_HEADER = ('Key name', 'Last updated')
ACL_HEADER = ('Principal', 'Permission')
DASH_MARKER = '# ' + '-' * 70 + '\n'


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help="Creates a secret scope.")
@click.option('--scope', required=True, type=SecretScopeClickType(), help=SecretScopeClickType.help)
@click.option('--initial-manage-principal',
              help='The initial principal that can manage the created secret scope.'
              ' If specified, the initial ACL with MANAGE permission applied to the scope is'
              ' assigned to the supplied principal (user or group). Currently, the only supported'
              ' principal for this option is the group "users", which contains all users in the'
              ' workspace. If not specified, the initial ACL with MANAGE permission applied to the'
              ' scope is assigned to the request issuer\'s user identity.')
@click.option('--scope-backend-type',
              type=click.Choice(['AZURE_KEYVAULT', 'DATABRICKS']),
              default='DATABRICKS', help='The backend that will be used for this secret scope. '
                                         'Options are (case-sensitive): 1) \'AZURE_KEYVAULT\' and '
                                         '2) \'DATABRICKS\' (default option)'
                                         '\nNote: To create an Azure Keyvault, be sure '
                                         'to configure an AAD Token using '
                                         '\'databricks configure --aad-token\'')
@click.option('--resource-id', default=None, type=click.STRING,
              help='The resource ID associated with the azure keyvault to be used as the backend'
                   ' for the secret scope. NOTE: Only use with azure-keyvault as backend')
@click.option('--dns-name', default=None, type=click.STRING,
              help='The dns name associated with the azure keyvault to be used as the'
                   ' backed for the secret scope. NOTE: Only use with azure-keyvault as backend')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_scope(api_client, scope, initial_manage_principal,
                 scope_backend_type, resource_id, dns_name):
    """
    Creates a new secret scope with given name.
    """
    backend_azure_keyvault = None

    if resource_id is not None and dns_name is not None:
        backend_azure_keyvault = {
            'resource_id': resource_id,
            'dns_name': dns_name
        }
    SecretApi(api_client).create_scope(scope, initial_manage_principal,
                                       scope_backend_type, backend_azure_keyvault)


def _scopes_to_table(scopes_json):
    ret = []
    for s in scopes_json.get('scopes', []):
        if "keyvault_metadata" in s:
            url = s["keyvault_metadata"]["dns_name"]
            ret.append((truncate_string(s['name']), s['backend_type'], url))
        else:
            ret.append((truncate_string(s['name']), s['backend_type'], "N/A"))
    return ret


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists all secret scopes.')
@click.option('--output', help=OutputClickType.help, type=OutputClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_scopes(api_client, output):
    """
    Lists all secret scopes.
    """
    scopes_json = SecretApi(api_client).list_scopes()
    if OutputClickType.is_json(output):
        click.echo(pretty_format(scopes_json))
    else:
        click.echo(tabulate(_scopes_to_table(scopes_json), headers=SCOPE_HEADER))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Deletes a secret scope.')
@click.option('--scope', required=True, type=SecretScopeClickType(), help=SecretScopeClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_scope(api_client, scope):
    """
    Deletes a secret scope.
    """
    SecretApi(api_client).delete_scope(scope)


def _verify_and_translate_options(string_value, binary_file):
    """
    Translates options into actual parameters for API call.
    Return tuple with two values representing (string_value, bytes_value).
    """
    if string_value and binary_file:
        error_and_quit("At most one of {} should be provided."
                       .format(['string-value', 'binary-file']))

    elif string_value is None and binary_file is None:
        prompt = '# Do not edit the above line. Everything below it will be ignored.\n' + \
            '# Please input your secret value above the line. Text will be stored in\n' + \
            '# UTF-8 (MB4) form and any trailing new line will be stripped.\n' + \
            '# Exit without saving will abort writing secret.'

        # underlying edit function made sure using a temporary file for editing
        content = click.edit('\n\n' + DASH_MARKER + prompt)
        # return None means editor is closed without changes
        if content is None:
            error_and_quit('No changes made, write secret aborted.'
                           ' Please follow the instruction to input secret value.')

        elif DASH_MARKER not in content:
            error_and_quit('Please DO NOT edit the line with dashes. Write secret aborted.')

        return content.split(DASH_MARKER, 1)[0].rstrip('\n'), None

    elif string_value is not None:
        return string_value, None

    elif binary_file is not None:
        with open(binary_file, 'rb') as f:
            binary_content = f.read()

        base64_bytes = base64.b64encode(binary_content)
        base64_str = base64_bytes.decode('utf-8')

        return None, base64_str


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Puts a secret in a scope. "write" is an alias for "put".')
@click.option('--scope', required=True, type=SecretScopeClickType(), help=SecretScopeClickType.help)
@click.option('--key', required=True, type=SecretKeyClickType(), help=SecretKeyClickType.help)
@click.option('--string-value', default=None,
              help='Read value from string and stored in UTF-8 (MB4) form')
@click.option('--binary-file', default=None, type=click.Path(exists=True, readable=True),
              help='Read value from binary-file and stored as bytes.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def put_secret(api_client, scope, key, string_value, binary_file):
    """
    Puts a secret in the provided scope with the given name.
    Overwrites any existing value if the name exists.

    You should specify at most one option in "string-value" and "binary-file".

    If "string-value", the argument will be stored in UTF-8 (MB4) form.

    If "binary-file", the argument should be a path to file. File content will be read as secret
    value and stored as bytes.

    If none of "string-value" and "binary-file" specified, an editor will be opened for
    inputting secret value. The value will be stored in UTF-8 (MB4) form.

    "databricks secrets write" is an alias for "databricks secrets put", and will be
    deprecated in a future release.
    """
    string_param, bytes_param = _verify_and_translate_options(string_value, binary_file)
    SecretApi(api_client).put_secret(scope, key, string_param, bytes_param)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Deletes a secret.')
@click.option('--scope', required=True, type=SecretScopeClickType(), help=SecretScopeClickType.help)
@click.option('--key', required=True, type=SecretKeyClickType(), help=SecretKeyClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_secret(api_client, scope, key):
    """
    Deletes the secret stored in this scope.
    """
    SecretApi(api_client).delete_secret(scope, key)


def _secrets_to_table(secrets_json):
    ret = []
    for s in secrets_json.get('secrets', []):
        ret.append((s['key'], s.get('last_updated_timestamp', 'Not Available')))
    return ret


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists all the secrets in a scope.')
@click.option('--scope', required=True, type=SecretScopeClickType(), help=SecretScopeClickType.help)
@click.option('--output', help=OutputClickType.help, type=OutputClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_secrets(api_client, scope, output):
    """
    Lists the secret keys that are stored at this scope. Also lists the last updated timestamp
    (UNIX time in milliseconds) if available.
    """
    secrets_json = SecretApi(api_client).list_secrets(scope)
    if OutputClickType.is_json(output):
        click.echo(pretty_format(secrets_json))
    else:
        click.echo(tabulate(_secrets_to_table(secrets_json), headers=SECRET_HEADER))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Creates or overwrites an access control rule for a principal applied to '
                          'a given secret scope. "write-acl" is an alias for "put-acl".')
@click.option('--scope', required=True, type=SecretScopeClickType(), help=SecretScopeClickType.help)
@click.option('--principal', required=True, type=SecretPrincipalClickType(),
              help=SecretPrincipalClickType.help)
@click.option('--permission', type=click.Choice(['MANAGE', 'WRITE', 'READ']),
              required=True, help='The permission to apply.')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def put_acl(api_client, scope, principal, permission):
    """
    Creates or overwrites the ACL associated with the given principal (user or group) on the
    specified secret scope.

    "databricks secrets write-acl" is an alias for "databricks secrets put-acl",
    and will be deprecated in a future release.
    """
    SecretApi(api_client).put_acl(scope, principal, permission)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Deletes an access control rule for a principal.')
@click.option('--scope', required=True, type=SecretScopeClickType(),
              help=SecretScopeClickType.help)
@click.option('--principal', required=True, type=SecretPrincipalClickType(),
              help=SecretPrincipalClickType.help)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_acl(api_client, scope, principal):
    """
    Deletes the given ACL on the given secret scope.
    """
    SecretApi(api_client).delete_acl(scope, principal)


def _acls_to_table(acls_json):
    ret = []
    for s in acls_json.get('items', []):
        ret.append((s['principal'], s['permission'].upper()))
    return ret


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Lists all access control rules for a given secret scope.')
@click.option('--scope', required=True, type=SecretScopeClickType(), help=SecretScopeClickType.help)
@click.option('--output', help=OutputClickType.help, type=OutputClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_acls(api_client, scope, output):
    """
    Lists the ACLs set on the given secret scope.
    """
    acls_json = SecretApi(api_client).list_acls(scope)
    if OutputClickType.is_json(output):
        click.echo(pretty_format(acls_json))
    else:
        click.echo(tabulate(_acls_to_table(acls_json), headers=ACL_HEADER))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Gets the details for an access control rule.')
@click.option('--scope', required=True, type=SecretScopeClickType(),
              help=SecretScopeClickType.help)
@click.option('--principal', required=True, type=SecretPrincipalClickType(),
              help=SecretPrincipalClickType.help)
@click.option('--output', help=OutputClickType.help, type=OutputClickType())
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def get_acl(api_client, scope, principal, output):
    """
    Describes the details about the given ACL for the principal and secret scope.
    """
    acl_json = SecretApi(api_client).get_acl(scope, principal)
    if OutputClickType.is_json(output):
        click.echo(pretty_format(acl_json))
    else:
        acl_list = _acls_to_table({'items': [acl_json]})
        click.echo(tabulate(acl_list, headers=ACL_HEADER))


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='Utility to interact with Databricks secret API.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
@eat_exceptions
def secrets_group():  # pragma: no cover
    """
    Utility to interact with secret API.
    """
    pass


secrets_group.add_command(create_scope, name='create-scope')
secrets_group.add_command(list_scopes, name='list-scopes')
secrets_group.add_command(delete_scope, name='delete-scope')
secrets_group.add_command(put_secret, name='put')
secrets_group.add_command(put_secret, name='write')
secrets_group.add_command(delete_secret, name='delete')
secrets_group.add_command(list_secrets, name='list')
secrets_group.add_command(put_acl, name='put-acl')
secrets_group.add_command(put_acl, name='write-acl')
secrets_group.add_command(delete_acl, name='delete-acl')
secrets_group.add_command(list_acls, name='list-acls')
secrets_group.add_command(get_acl, name='get-acl')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Databricks Python REST Client 2.0 for interacting with various services.

Currently supports services including clusters, clusters policies and jobs.

Requires Python 2.7.9 or above.

To get started, below is an example usage of the Python API client.

  # Import databricks package:
  from databricks import *

  # Create a client:
  userName = "user@company.com"
  password = "MySecretPassword"
  client = ApiClient(userName, password, host = "https://dbc-12345678-9101.cloud.databricks.com")

  # List jobs:
  jobs = JobsService(client)
  print jobs.list_jobs()

  # For help:
  help(databricks)

  # To examine available services:
  help(databricks.service)

  # To examine the jobs API:
  help(JobsService)
"""
from .service import *
from .api_client import ApiClient



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from databricks_cli.sdk import JobsService


class JobsApi(object):
    def __init__(self, api_client):
        self.client = JobsService(api_client)

    def create_job(self, json, headers=None, version=None):
        return self.client.client.perform_query('POST', '/jobs/create', data=json, headers=headers,
                                                version=version)

    def list_jobs(self, job_type=None, expand_tasks=None, offset=None, limit=None, headers=None,
                  version=None, name=None):
        resp = self.client.list_jobs(job_type=job_type, expand_tasks=expand_tasks, offset=offset,
                                     limit=limit, headers=headers, version=version, 
                                     name=name)
        if 'jobs' not in resp:
            resp['jobs'] = []
        return resp

    def delete_job(self, job_id, headers=None, version=None):
        return self.client.delete_job(job_id, headers=headers, version=version)

    def get_job(self, job_id, headers=None, version=None):
        return self.client.get_job(job_id, headers=headers, version=version)

    def reset_job(self, json, headers=None, version=None):
        return self.client.client.perform_query('POST', '/jobs/reset', data=json, headers=headers,
                                                version=version)

    def run_now(self, job_id, jar_params, notebook_params, python_params, spark_submit_params,
                python_named_params=None, idempotency_token=None, headers=None, version=None):
        return self.client.run_now(job_id, jar_params, notebook_params, python_params,
                                   spark_submit_params, python_named_params,
                                   idempotency_token, headers=headers, version=version)

    def _list_jobs_by_name(self, name, headers=None):
        jobs = self.list_jobs(headers=headers, name=name)['jobs']
        result = list(filter(lambda job: job['settings']['name'] == name, jobs))
        return result



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from click import ParamType, Option, MissingParameter, UsageError


class OutputClickType(ParamType):
    name = 'FORMAT'
    help = 'can be "JSON" or "TABLE". Set to TABLE by default.'

    def convert(self, value, param, ctx):
        if value is None:
            return value
        if value.lower() != 'json' and value.lower() != 'table':
            raise RuntimeError('output must be "json" or "table"')
        return value

    @classmethod
    def is_json(cls, value):
        return value is not None and value.lower() == 'json'

    @classmethod
    def is_table(cls, value):
        return value is not None and value.lower() == 'table'


class JsonClickType(ParamType):
    name = 'JSON'

    @classmethod
    def help(cls, endpoint):
        return 'JSON string to POST to {}.'.format(endpoint)


class JobIdClickType(ParamType):
    name = 'JOB_ID'
    help = 'Can be found in the URL at https://*.cloud.databricks.com/#job/$JOB_ID.'


class RunIdClickType(ParamType):
    name = 'RUN_ID'


class ClusterIdClickType(ParamType):
    name = 'CLUSTER_ID'
    help = ('Can be found in the URL at '
            'https://*.cloud.databricks.com/#/setting/clusters/$CLUSTER_ID/configuration.')


class ClusterPolicyIdClickType(ParamType):
    name = 'POLICY_ID'
    help = ('Can be found in the URL at '
            'https://*.cloud.databricks.com/#/setting/clusters/cluster-policies/view/$POLICY_ID.')


class InstancePoolIdClickType(ParamType):
    name = 'INSTANCE_POOL_ID'
    help = ('Can be found in the URL at '
            'https://*.cloud.databricks.com/#setting/clusters/instance-pools/view/'
            '$INSTANCE_POOL_ID')


class SecretScopeClickType(ParamType):
    name = 'SCOPE'
    help = 'The name of the secret scope.'


class SecretKeyClickType(ParamType):
    name = 'KEY'
    help = 'The name of the secret key.'


class SecretPrincipalClickType(ParamType):
    name = 'PRINCIPAL'
    help = 'The name of the principal.'


class PipelineSpecClickType(ParamType):
    name = 'SPEC'
    help = '[Deprecated] Use the settings option instead. \n' + \
           'The path to the pipelines settings file.'


class PipelineSettingClickType(ParamType):
    name = 'SETTINGS'
    help = 'The path to the pipelines settings file.'


class PipelineIdClickType(ParamType):
    name = 'PIPELINE_ID'
    help = 'The pipeline ID.'


class MetastoreIdClickType(ParamType):
    name = 'METASTORE_ID'
    help = 'ID of the Metastore'


class WorkspaceIdClickType(ParamType):
    name = 'WORKSPACE_ID'
    help = 'ID of the Workspace'


class OneOfOption(Option):
    def __init__(self, *args, **kwargs):
        self.one_of = kwargs.pop('one_of')
        super(OneOfOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        cleaned_opts = set([o.replace('_', '-') for o in opts.keys()])
        if len(cleaned_opts.intersection(set(self.one_of))) == 0:
            raise MissingParameter('One of {} must be provided.'.format(self.one_of))
        if len(cleaned_opts.intersection(set(self.one_of))) > 1:
            raise UsageError('Only one of {} should be provided.'.format(self.one_of))
        return super(OneOfOption, self).handle_parse_result(ctx, opts, args)


class OptionalOneOfOption(Option):
    def __init__(self, *args, **kwargs):
        self.one_of = kwargs.pop('one_of')
        super(OptionalOneOfOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        cleaned_opts = set([o.replace('_', '-') for o in opts.keys()])
        if len(cleaned_opts.intersection(set(self.one_of))) > 1:
            raise UsageError('Only one of {} should be provided.'.format(self.one_of))
        return super(OptionalOneOfOption, self).handle_parse_result(ctx, opts, args)


class ContextObject(object):
    def __init__(self):
        self._profile = None
        self._debug = False

    def set_debug(self, debug=False):
        self._debug = debug

        if not self._debug:
            return

        # These two lines enable debugging at httplib level (requests->urllib3->http.client)
        # You will see the REQUEST, including HEADERS and DATA,
        # and RESPONSE with HEADERS but without DATA.
        # The only thing missing will be the response.body which is not logged.
        try:
            import http.client as http_client
        except ImportError:
            # Python 2
            import httplib as http_client

        click.echo("HTTP debugging enabled")
        http_client.HTTPConnection.debuglevel = 1

    @property
    def debug_mode(self):
        return self._debug

    def set_profile(self, profile):
        if self._profile is not None:
            raise UsageError('--profile can only be provided once. '
                             'The profiles [{}, {}] were provided.'.format(self._profile, profile))
        self._profile = profile

    def get_profile(self):
        return self._profile


class RequiredOptions(Option):
    def __init__(self, *args, **kwargs):
        self.one_of = kwargs.pop('one_of')
        super(RequiredOptions, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        cleaned_opts = set([o.replace('_', '-') for o in opts.keys()])
        if len(cleaned_opts.intersection(set(self.one_of))) == 0:
            raise MissingParameter('One of {} must be provided.'.format(self.one_of))
        if len(cleaned_opts.intersection(set(self.one_of))) > 1:
            raise UsageError('Only one of {} should be provided.'.format(self.one_of))
        return super(RequiredOptions, self).handle_parse_result(ctx, opts, args)



# Databricks CLI
# Copyright 2021 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import base64
import hashlib
import json
import os
import webbrowser

from datetime import datetime, timedelta, tzinfo

import click

import jwt
from jwt import PyJWTError

import oauthlib.oauth2
from oauthlib.oauth2.rfc6749.errors import OAuth2Error

import requests
from requests.exceptions import RequestException

from databricks_cli.utils import error_and_quit

try:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
except ImportError:
    from http.server import BaseHTTPRequestHandler, HTTPServer


# This could use 'import secrets' in Python 3
def token_urlsafe(nbytes=32):
    tok = os.urandom(nbytes)
    return base64.urlsafe_b64encode(tok).rstrip(b'=').decode('ascii')


# This could be datetime.timezone.utc in Python 3
class UTCTimeZone(tzinfo):
    """UTC"""
    def utcoffset(self, dt):
        #pylint: disable=unused-argument
        return timedelta(0)

    def tzname(self, dt):
        #pylint: disable=unused-argument
        return "UTC"

    def dst(self, dt):
        #pylint: disable=unused-argument
        return timedelta(0)


# Some contant values
OIDC_REDIRECTOR_PATH = "oidc"
CLIENT_ID = "databricks-cli"
REDIRECT_PORT = 8020
UTC = UTCTimeZone()


def get_client(client_id=CLIENT_ID):
    return oauthlib.oauth2.WebApplicationClient(client_id)


def get_redirect_url(port=REDIRECT_PORT):
    return "http://localhost:{port}".format(port=port)


def fetch_well_known_config(idp_url):
    known_config_url = "{idp_url}/.well-known/oauth-authorization-server".format(idp_url=idp_url)
    try:
        response = requests.request(method="GET", url=known_config_url)
    except RequestException:
        error_and_quit("Unable to fetch OAuth configuration from {idp_url}.\n"
                       "Verify it is a valid workspace URL and that OAuth is "
                       "enabled on this account.".format(idp_url=idp_url))

    if response.status_code != 200:
        error_and_quit("Received status {status} OAuth configuration from "
                       "{idp_url}.\n Verify it is a valid workspace URL and "
                       "that OAuth is enabled on this account."
                       .format(status=response.status_code, idp_url=idp_url))
    try:
        return json.loads(response.text)
    except json.decoder.JSONDecodeError:
        error_and_quit("Unable to decode OAuth configuration from {idp_url}.\n"
                       "Verify it is a valid workspace URL and that OAuth is "
                       "enabled on this account.".format(idp_url=idp_url))


def get_idp_url(host):
    maybe_scheme = "https://" if not host.startswith("https://") else ""
    maybe_trailing_slash = "/" if not host.endswith("/") else ""
    return "{scheme}{host}{trailing}{path}".format(
        scheme=maybe_scheme, host=host, trailing=maybe_trailing_slash, path=OIDC_REDIRECTOR_PATH)


def get_challenge(verifier_string=token_urlsafe(32)):
    digest = hashlib.sha256(verifier_string.encode('UTF-8')).digest()
    challenge_string = base64.urlsafe_b64encode(digest).decode("UTF-8").replace('=', '')
    return verifier_string, challenge_string


# This is a janky global that is used to store the path of the single request the HTTP server
# will receive.
global_request_path = None


def set_request_path(path):
    global global_request_path
    global_request_path = path


class SingleRequestHandler(BaseHTTPRequestHandler):
    RESPONSE_BODY = """<html>
<head>
  <title>Close this Tab</title>
  <style>
    body {
      font-family: "Barlow", Helvetica, Arial, sans-serif;
      padding: 20px;
      background-color: #f3f3f3;
    }
  </style>
</head>
<body>
  <h1>Please close this tab.</h1>
  <p>
    The Databricks CLI received a response. You may close this tab.
  </p>
</body>
</html>""".encode("utf-8")

    def do_GET(self):  # nopep8
        self.send_response(200, "Success")
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(self.RESPONSE_BODY)
        set_request_path(self.path)

    def log_message(self, format, *args):
        #pylint: disable=redefined-builtin
        #pylint: disable=unused-argument
        return


def get_authorization_code(client, auth_url, redirect_url, scope, state, challenge, port):
    #pylint: disable=unused-variable
    (auth_req_uri, headers, body) = client.prepare_authorization_request(
        authorization_url=auth_url,
        redirect_url=redirect_url,
        scope=scope,
        state=state,
        code_challenge=challenge,
        code_challenge_method="S256")
    click.echo("Opening {uri}".format(uri=auth_req_uri))

    with HTTPServer(("", port), SingleRequestHandler) as httpd:
        webbrowser.open_new(auth_req_uri)
        click.echo("Listening for OAuth authorization callback at {uri}"
                   .format(uri=redirect_url))
        httpd.handle_request()

    if not global_request_path:
        error_and_quit("No path parameters were returned to the callback at {uri}"
                       .format(uri=redirect_url))
    # This is a kludge because the parsing library expects https callbacks
    # We should probably set it up using https
    full_redirect_url = "https://localhost:{port}/{path}".format(
        port=port, path=global_request_path)
    try:
        authorization_code_response = \
            client.parse_request_uri_response(full_redirect_url, state=state)
    except OAuth2Error as err:
        error_and_quit("OAuth Token Request error {error}".format(error=err.description))
    return authorization_code_response


def send_auth_code_token_request(client, token_request_url, redirect_url, code, verifier):
    token_request_body = client.prepare_request_body(code=code, redirect_uri=redirect_url)
    data = "{body}&code_verifier={verifier}".format(body=token_request_body, verifier=verifier)
    return send_token_request(token_request_url, data)


def send_token_request(token_request_url, data):
    headers = {
        "Accept": "application/json",
        "Content-Type": "application/x-www-form-urlencoded"
    }
    response = requests.request(method="POST", url=token_request_url, data=data, headers=headers)
    oauth_response = json.loads(response.text)
    return oauth_response


def send_refresh_token_request(hostname, refresh_token):
    idp_url = get_idp_url(hostname)
    oauth_config = fetch_well_known_config(idp_url)
    token_request_url = oauth_config['token_endpoint']
    client = get_client()
    token_request_body = client.prepare_refresh_body(
        refresh_token=refresh_token, client_id=client.client_id)
    return send_token_request(token_request_url, token_request_body)


def get_tokens_from_response(oauth_response):
    access_token = oauth_response['access_token']
    refresh_token = oauth_response['refresh_token'] if 'refresh_token' in oauth_response else None
    return access_token, refresh_token


def check_and_refresh_access_token(hostname, access_token, refresh_token):
    now = datetime.now(tz=UTC)
    # If we can't decode an expiration time, this will be expired by default.
    expiration_time = now
    try:
        # This token has already been verified and we are just parsing it.
        # If it has been tampered with, it will be rejected on the server side.
        # This avoids having to fetch the public key from the issuer and perform
        # an unnecessary signature verification.
        decoded = jwt.decode(access_token, options={"verify_signature": False})
        expiration_time = datetime.fromtimestamp(decoded['exp'], tz=UTC)
    except PyJWTError as err:
        error_and_quit(err)

    if expiration_time > now:
        # The access token is fine. Just return it.
        return access_token, refresh_token, False

    if not refresh_token:
        error_and_quit("OAuth access token expired on {expiration_time}."
                       .format(expiration_time=expiration_time))

    # Try to refresh using the refresh token
    click.echo("Attempting to refresh OAuth access token that expired on {expiration_time}"
               .format(expiration_time=expiration_time))
    oauth_response = send_refresh_token_request(hostname, refresh_token)
    fresh_access_token, fresh_refresh_token = get_tokens_from_response(oauth_response)
    return fresh_access_token, fresh_refresh_token, True


def get_tokens(hostname, scope=None):
    idp_url = get_idp_url(hostname)
    oauth_config = fetch_well_known_config(idp_url)
    # We are going to override oauth_config["authorization_endpoint"] use the
    # /oidc redirector on the hostname, which may inject additional parameters.
    auth_url = "{}/v1/authorize".format(get_idp_url(hostname))
    state = token_urlsafe(16)
    (verifier, challenge) = get_challenge()
    client = get_client()
    redirect_url = get_redirect_url()
    try:
        auth_response = get_authorization_code(
            client,
            auth_url,
            redirect_url,
            scope,
            state,
            challenge,
            REDIRECT_PORT)
    except OAuth2Error as err:
        error_and_quit("OAuth Authorization Error: {error}".format(error=err.description))

    token_request_url = oauth_config["token_endpoint"]
    code = auth_response['code']
    oauth_response = \
        send_auth_code_token_request(client, token_request_url, redirect_url, code, verifier)
    return get_tokens_from_response(oauth_response)



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from databricks_cli.sdk import ClusterService


class ClusterApi(object):
    def __init__(self, api_client):
        self.client = ClusterService(api_client)

    def create_cluster(self, json):
        return self.client.client.perform_query('POST', '/clusters/create', data=json)

    def edit_cluster(self, json):
        return self.client.client.perform_query('POST', '/clusters/edit', data=json)

    def start_cluster(self, cluster_id):
        return self.client.start_cluster(cluster_id)

    def restart_cluster(self, cluster_id):
        return self.client.restart_cluster(cluster_id)

    def resize_cluster(self, cluster_id, num_workers):
        return self.client.resize_cluster(cluster_id, num_workers=num_workers)

    def delete_cluster(self, cluster_id):
        return self.client.delete_cluster(cluster_id)

    def get_cluster(self, cluster_id):
        return self.client.get_cluster(cluster_id)

    def list_clusters(self):
        return self.client.list_clusters()

    def list_zones(self):
        return self.client.list_available_zones()

    def list_node_types(self):
        return self.client.list_node_types()

    def spark_versions(self):
        return self.client.list_spark_versions()

    def permanent_delete(self, cluster_id):
        return self.client.permanent_delete_cluster(cluster_id)

    def get_cluster_ids_by_name(self, cluster_name):
        data = self.client.list_clusters()
        return [c for c in data.get('clusters', []) if c.get('cluster_name') == cluster_name]

    def get_cluster_id_for_name(self, cluster_name):
        """
        Given a cluster name, this will return a single cluster id for that name.
        If there are multiple clusters with the same name it will raise a RuntimeError.
        If there are no clusters with the name it will raise a RuntimeError.
        """
        clusters_by_name = self.get_cluster_ids_by_name(cluster_name)
        cluster_ids = [
            cluster['cluster_id'] for cluster in clusters_by_name if
            cluster and 'cluster_id' in cluster
        ]

        if len(cluster_ids) == 0:
            raise RuntimeError('No clusters with name {} were found'.format(cluster_name))

        if len(cluster_ids) > 1:
            raise RuntimeError('More than 1 cluster was named {}, '.format(cluster_name) +
                               'please use --cluster-id.\n' +
                               'Cluster ids found: {}'.format(', '.join(cluster_ids))
                               )
        return cluster_ids[0]

    def get_cluster_by_name(self, cluster_name):
        """
        Given a cluster name, this will return the cluster config for that cluster.
        If there are multiple clusters with the same name it will raise a RuntimeError.
        If there are no clusters with the name it will raise a RuntimeError.
        """
        cluster_id = self.get_cluster_id_for_name(cluster_name)
        return self.get_cluster(cluster_id)

    def get_events(self, cluster_id, start_time, end_time, order, event_types, offset, limit):
        return self.client.get_events(cluster_id, start_time, end_time, order, event_types,
                                      offset, limit)



# Databricks CLI
# Copyright 2022 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from databricks_cli.unity_catalog.utils import mc_pretty_format


class UnityCatalogService(object):
    def __init__(self, client):
        self.client = client

    # Metastore Operations

    def create_metastore(self, name, storage_root, region, headers=None):
        _data = {
            'name': name,
            'storage_root': storage_root,
        }
        if region is not None:
            _data['region'] = region
        return self.client.perform_query('POST', '/unity-catalog/metastores', data=_data,
                                         headers=headers)

    def list_metastores(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/metastores', data=_data,
                                         headers=headers)

    def get_metastore(self, metastore_id, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/metastores/%s' % (metastore_id),
                                         data=_data, headers=headers)

    def get_metastore_summary(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/metastore_summary', data=_data,
                                         headers=headers)

    def update_metastore(self, metastore_id, metastore_spec, headers=None):
        return self.client.perform_query('PATCH', '/unity-catalog/metastores/%s' % (metastore_id),
                                         data=metastore_spec, headers=headers)

    def delete_metastore(self, metastore_id, force=None, headers=None):
        _data = {}
        if force is not None:
            _data['force'] = force

        return self.client.perform_query('DELETE', '/unity-catalog/metastores/%s' % (metastore_id),
                                         data=_data, headers=headers)

    def create_metastore_assignment(self, workspace_id, metastore_id, default_catalog_name=None,
                                    headers=None):
        _data = {
            'metastore_id': metastore_id
        }
        if default_catalog_name is not None:
            _data['default_catalog_name'] = default_catalog_name
        url = '/unity-catalog/workspaces/%s/metastore' % (workspace_id)
        return self.client.perform_query('PUT', url, data=_data, headers=headers)

    def update_metastore_assignment(self, workspace_id, metastore_id, default_catalog_name,
                                    headers=None):
        _data = {
            'metastore_id': metastore_id,
            'default_catalog_name': default_catalog_name
        }
        url = '/unity-catalog/workspaces/%s/metastore' % (workspace_id)
        return self.client.perform_query('PATCH', url, data=_data, headers=headers)

    def delete_metastore_assignment(self, workspace_id, metastore_id, headers=None):
        _data = {
            'metastore_id': metastore_id
        }
        url = '/unity-catalog/workspaces/%s/metastore' % (workspace_id)
        return self.client.perform_query('DELETE', url, data=_data, headers=headers)

    def get_current_metastore_assignment(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/current-metastore-assignment',
                                         data=_data, headers=headers)

    # External Location Operations

    def create_external_location(self, loc_spec, skip_validation, headers=None):
        # Merge the skip_validation arg, since it's not a query arg and the
        # ExternalLocationInfo spec is 'inline'
        if skip_validation:
            loc_spec['skip_validation'] = skip_validation
        url = '/unity-catalog/external-locations'
        return self.client.perform_query('POST', url, data=loc_spec, headers=headers)

    def list_external_locations(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/external-locations', data=_data,
                                         headers=headers)

    def get_external_location(self, name, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/external-locations/%s' % (name),
                                         data=_data, headers=headers)

    def update_external_location(self, name, loc_spec, force, skip_validation, headers=None):
        _data = loc_spec
        # Merge the skip_validation arg, since it's not a query arg and the
        # ExternalLocationInfo spec is 'inline'
        if skip_validation:
            _data['skip_validation'] = skip_validation
        # Same for the 'force' field.
        if force:
            _data["force"] = True

        return self.client.perform_query('PATCH', '/unity-catalog/external-locations/%s' % (name),
                                         data=_data, headers=headers)

    def delete_external_location(self, name, force, headers=None):
        _data = {
            "force": force
        }
        return self.client.perform_query('DELETE', '/unity-catalog/external-locations/%s' % (name),
                                         data=_data, headers=headers)

    def validate_external_location(self, validation_spec, headers=None):
        return self.client.perform_query('POST', '/unity-catalog/validate-storage-credentials',
                                         data=validation_spec, headers=headers)

    # Data Access Configuration Operations

    def create_dac(self, metastore_id, dac_spec, skip_validation, headers=None):
        if skip_validation:
            dac_spec['skip_validation'] = skip_validation
        url = '/unity-catalog/metastores/%s/data-access-configurations' % (metastore_id)
        return self.client.perform_query('POST', url, data=dac_spec, headers=headers)

    def list_dacs(self, metastore_id, headers=None):
        _data = {}
        url = '/unity-catalog/metastores/%s/data-access-configurations' % (metastore_id)
        return self.client.perform_query('GET', url, data=_data, headers=headers)

    def get_dac(self, metastore_id, dac_id, headers=None):
        url = '/unity-catalog/metastores/%s/data-access-configurations/%s' % (metastore_id, dac_id)
        return self.client.perform_query('GET', url, headers=headers)

    def delete_dac(self, metastore_id, dac_id, headers=None):
        url = '/unity-catalog/metastores/%s/data-access-configurations/%s' % (metastore_id, dac_id)
        return self.client.perform_query('DELETE', url, headers=headers)

    # Storage Credential Operations

    def create_storage_credential(self, cred_spec, skip_validation, headers=None):
        # Merge the skip_validation arg, since it's not a query arg and the
        # StorageCredentialInfo spec is 'inline'
        if skip_validation:
            cred_spec['skip_validation'] = skip_validation
        url = '/unity-catalog/storage-credentials'
        return self.client.perform_query('POST', url, data=cred_spec, headers=headers)

    def list_storage_credentials(self, name_pattern=None, headers=None):
        _data = {}
        if name_pattern is not None:
            _data['name_pattern'] = name_pattern

        return self.client.perform_query('GET', '/unity-catalog/storage-credentials',
                                         data=_data, headers=headers)

    def get_storage_credential(self, name, headers=None):
        _data = {}

        return self.client.perform_query('GET', '/unity-catalog/storage-credentials/%s' % (name),
                                         data=_data, headers=headers)

    def update_storage_credential(self, name, cred_spec, skip_validation, headers=None):
        # Merge the skip_validation arg, since it's not a query arg and the
        # StorageCredentialInfo spec is 'inline'
        if skip_validation:
            cred_spec['skip_validation'] = skip_validation
        return self.client.perform_query('PATCH', '/unity-catalog/storage-credentials/%s' % (name),
                                         data=cred_spec, headers=headers)

    def delete_storage_credential(self, name, force, headers=None):
        _data = {}
        if force:
            _data["force"] = True

        return self.client.perform_query('DELETE', '/unity-catalog/storage-credentials/%s' % (name),
                                         data=_data, headers=headers)

    # Catalog Operations

    def create_catalog(self, name, comment=None, provider=None, share=None, headers=None):
        _data = {
            'name': name,
        }
        if comment is not None:
            _data['comment'] = comment
        if provider is not None:
            _data['provider_name'] = provider
        if share is not None:
            _data['share_name'] = share
        return self.client.perform_query('POST', '/unity-catalog/catalogs', data=_data,
                                         headers=headers)

    def list_catalogs(self, headers=None):
        _data = {}

        return self.client.perform_query('GET', '/unity-catalog/catalogs', data=_data,
                                         headers=headers)

    def get_catalog(self, name, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/catalogs/%s' % (name),
                                         data=_data, headers=headers)

    def update_catalog(self, name, catalog_spec, headers=None):
        return self.client.perform_query('PATCH', '/unity-catalog/catalogs/%s' % (name),
                                         data=catalog_spec, headers=headers)

    def delete_catalog(self, name, headers=None):
        _data = {}
        return self.client.perform_query('DELETE', '/unity-catalog/catalogs/%s' % (name),
                                         data=_data, headers=headers)

    # Schema Operations

    def create_schema(self, catalog_name, new_schema_name, comment=None, headers=None):
        _data = {
            'catalog_name': catalog_name,
            'name': new_schema_name,
        }
        if comment is not None:
            _data['comment'] = comment
        return self.client.perform_query('POST', '/unity-catalog/schemas', data=_data,
                                         headers=headers)

    def list_schemas(self, catalog_name=None, name_regex=None, headers=None):
        _data = {}
        if catalog_name is not None:
            _data['catalog_name'] = catalog_name
        if name_regex is not None:
            _data['schema_name_regex'] = name_regex

        return self.client.perform_query('GET', '/unity-catalog/schemas', data=_data,
                                         headers=headers)

    def list_lineages_by_table(self, table_name=None, headers=None):
        """
        List table lineage by table name
        """
        _data = {}
        if table_name is not None:
            _data['table_name'] = table_name

        return self.client.perform_query('GET', '/lineage-tracking/table-lineage/get', data=_data,
                                         headers=headers)

    def list_lineages_by_column(self, table_name=None, column_name=None, headers=None):
        """
        List column lineage by table name and comlumn name
        """
        _data = {}
        if table_name is not None:
            _data['table_name'] = table_name
        if column_name is not None:
            _data['column_name'] = column_name

        return self.client.perform_query('GET', '/lineage-tracking/column-lineage/get', data=_data,
                                         headers=headers)

    def get_schema(self, full_name, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/schemas/%s' % (full_name),
                                         data=_data, headers=headers)

    def update_schema(self, full_name, schema_spec, headers=None):
        return self.client.perform_query('PATCH', '/unity-catalog/schemas/%s' % (full_name),
                                         data=schema_spec, headers=headers)

    def delete_schema(self, full_name, headers=None):
        _data = {}
        return self.client.perform_query('DELETE', '/unity-catalog/schemas/%s' % (full_name),
                                         data=_data, headers=headers)

    # Table Operations

    def create_table(self, table_spec, headers=None):
        return self.client.perform_query('POST', '/unity-catalog/tables', data=table_spec,
                                         headers=headers)

    def list_tables(self, catalog_name, schema_name=None, name_regex=None, headers=None):
        _data = {
            'catalog_name': catalog_name
        }
        if schema_name is not None:
            _data['schema_name'] = schema_name
        if name_regex is not None:
            _data['table_name_regex'] = name_regex

        return self.client.perform_query('GET', '/unity-catalog/tables', data=_data,
                                         headers=headers)

    def list_table_summaries(self, catalog_name, headers=None):
        _data = {
            'catalog_name': catalog_name
        }
        return self.client.perform_query('GET', '/unity-catalog/table-summaries', data=_data,
                                         headers=headers)

    def get_table(self, full_name, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/tables/%s' % (full_name),
                                         data=_data, headers=headers)

    def update_table(self, full_name, table_spec, headers=None):
        return self.client.perform_query('PATCH', '/unity-catalog/tables/%s' % (full_name),
                                         data=table_spec, headers=headers)

    def delete_table(self, full_name, headers=None):
        _data = {}
        return self.client.perform_query('DELETE', '/unity-catalog/tables/%s' % (full_name),
                                         data=_data, headers=headers)

    # Share Operations

    def create_share(self, name, headers=None):
        _data = {
            'name': name
        }
        return self.client.perform_query('POST', '/unity-catalog/shares', data=_data,
                                         headers=headers)

    def list_shares(self, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/shares', data=_data,
                                         headers=headers)

    def get_share(self, name, include_shared_data, headers=None):
        _data = {'include_shared_data': include_shared_data}

        return self.client.perform_query('GET', '/unity-catalog/shares/%s' % (name),
                                         data=_data, headers=headers)

    def update_share(self, name, share_spec, headers=None):
        return self.client.perform_query('PATCH', '/unity-catalog/shares/%s' % (name),
                                         data=share_spec, headers=headers)

    def delete_share(self, name, headers=None):
        _data = {}
        return self.client.perform_query('DELETE', '/unity-catalog/shares/%s' % (name),
                                         data=_data, headers=headers)

    def list_share_permissions(self, name, headers=None):
        _data = {}
        return self.client.perform_query('GET', '/unity-catalog/shares/%s/permissions' % (name),
                                         data=_data, headers=headers)

    def update_share_permissions(self, name, perm_spec, headers=None):
        return self.client.perform_query('PATCH', '/unity-catalog/shares/%s/permissions' % (name),
                                         data=perm_spec, headers=headers)

    # Recipient Operations

    def create_recipient(self, name, comment=None, sharing_id=None,
                         allowed_ip_addresses=None, custom_properties=None, headers=None):
        _data = {
            'name': name,
        }
        if comment is not None:
            _data['comment'] = comment
        if sharing_id is not None:
            _data['data_recipient_global_metastore_id'] = sharing_id
            _data['authentication_type'] = 'DATABRICKS'
        else:
            _data['authentication_type'] = 'TOKEN'
        if allowed_ip_addresses is not None:
            _data['ip_access_list'] = {
                'allowed_ip_addresses': allowed_ip_addresses,
            }
        if custom_properties is not None:
            _data['properties_kvpairs'] = {
                'properties': custom_properties
            }

        return self.client.perform_query('POST', '/unity-catalog/recipients', data=_data,
                                         headers=headers)

    def list_recipients(self, headers=None):
        _data = {}

        return self.client.perform_query('GET', '/unity-catalog/recipients', data=_data,
                                         headers=headers)

    def get_recipient(self, name, headers=None):
        _data = {}

        return self.client.perform_query('GET', '/unity-catalog/recipients/%s' % (name),
                                         data=_data, headers=headers)

    def update_recipient(self, name, recipient_spec, headers=None):
        return self.client.perform_query('PATCH', '/unity-catalog/recipients/%s' % (name),
                                         data=recipient_spec, headers=headers)

    def rotate_recipient_token(self, name, existing_token_expire_in_seconds=None, headers=None):
        _data = {
            'name': name,
        }
        if existing_token_expire_in_seconds is not None:
            _data['existing_token_expire_in_seconds'] = existing_token_expire_in_seconds
        return self.client.perform_query('POST', '/unity-catalog/recipients/%s/rotate-token' %
                                         (name), data=_data, headers=headers)

    def get_recipient_share_permissions(self, name, headers=None):
        _data = {}

        return self.client.perform_query('GET', '/unity-catalog/recipients/%s/share-permissions'
                                         % (name), data=_data, headers=headers)

    def delete_recipient(self, name, headers=None):
        _data = {}

        return self.client.perform_query('DELETE', '/unity-catalog/recipients/%s' % (name),
                                         data=_data, headers=headers)

    # Provider Operations

    def create_provider(self, name, comment, recipient_profile=None, headers=None):
        _data = {
            'name': name,
        }
        if comment is not None:
            _data['comment'] = comment
        if recipient_profile is not None:
            _data['recipient_profile_str'] = mc_pretty_format(recipient_profile)
            _data['authentication_type'] = 'TOKEN'
        else:
            _data['authentication_type'] = 'DATABRICKS'
        return self.client.perform_query('POST', '/unity-catalog/providers/',
                                         data=_data, headers=headers)

    def list_providers(self, headers=None):
        return self.client.perform_query('GET', '/unity-catalog/providers', data={},
                                         headers=headers)

    def get_provider(self, name, headers=None):
        return self.client.perform_query('GET', '/unity-catalog/providers/%s' % (name),
                                         data={}, headers=headers)

    def update_provider(self, name, provider_spec, headers=None):
        return self.client.perform_query('PATCH', '/unity-catalog/providers/%s' % (name),
                                         data=provider_spec, headers=headers)

    def delete_provider(self, name, headers=None):
        return self.client.perform_query('DELETE', '/unity-catalog/providers/%s' % (name),
                                         data={}, headers=headers)

    def list_provider_shares(self, name, headers=None):
        return self.client.perform_query('GET', '/unity-catalog/providers/%s/shares' % (name),
                                         data={}, headers=headers)

    # Permissions Operations

    def _permissions_url(self, sec_type, sec_name, effective=False):
        if effective:
            return '/unity-catalog/effective-permissions/%s/%s' % (sec_type, sec_name)
        else:
            return '/unity-catalog/permissions/%s/%s' % (sec_type, sec_name)

    def get_permissions(self, sec_type, sec_name, headers=None):
        _data = {}
        return self.client.perform_query('GET', self._permissions_url(sec_type, sec_name),
                                         data=_data, headers=headers)

    def get_effective_permissions(self, sec_type, sec_name, headers=None):
        _data = {}
        return self.client.perform_query('GET', self._permissions_url(sec_type, sec_name,
                                                                      effective=True),
                                         data=_data, headers=headers)

    def update_permissions(self, sec_type, sec_name, perm_diff_spec, headers=None):
        _data = perm_diff_spec
        return self.client.perform_query('PATCH', self._permissions_url(sec_type, sec_name),
                                         data=_data, headers=headers)

    def replace_permissions(self, sec_type, sec_name, perm_spec, headers=None):
        _data = perm_spec
        return self.client.perform_query('PUT', self._permissions_url(sec_type, sec_name),
                                         data=_data, headers=headers)



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import click
from tabulate import tabulate

from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS
from databricks_cli.version import print_version_callback, version
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.workspace.api import WorkspaceApi
from databricks_cli.workspace.types import LanguageClickType, FormatClickType, WorkspaceFormat, \
    WorkspaceLanguage


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='List objects in the Databricks Workspace. ls and list are synonyms.')
@click.option('--absolute', is_flag=True, default=False,
              help='Displays absolute paths.')
@click.option('-l', '--long', 'is_long_form', is_flag=True, default=False,
              help='Displays full information including ObjectType, Path, Language')
@click.option('-i', '--id', 'with_object_id', is_flag=True, default=False,
              help='Displays ObjectId')
@click.argument('workspace_path', type=str, nargs=-1)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def ls_cli(api_client, is_long_form, with_object_id, absolute, workspace_path): # noqa
    """
    List objects in the Databricks Workspace.
    """
    if len(workspace_path) == 0:
        workspace_path = '/'
    else:
        workspace_path = workspace_path[0]
    objects = WorkspaceApi(api_client).list_objects(workspace_path)
    objects_table = [
        obj.to_row(is_long_form=is_long_form, is_absolute=absolute, with_object_id=with_object_id)
        for obj in objects
    ]
    table = tabulate(objects_table, tablefmt='plain')
    click.echo(table)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Make directories in the Databricks Workspace.')
@click.argument('workspace_path')
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def mkdirs_cli(api_client, workspace_path):
    """
    Make directories in the Databricks Workspace.

    Mkdirs will create directories along the path to the argument directory.
    """
    WorkspaceApi(api_client).mkdirs(workspace_path)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Imports a file from local to the Databricks workspace.')
@click.argument('source_path')
@click.argument('target_path')
@click.option('--language', '-l', required=True, type=LanguageClickType(),
              help=', '.join(WorkspaceLanguage.ALL))
@click.option('--format', '-f', default=WorkspaceFormat.SOURCE, type=FormatClickType())
@click.option('--overwrite', '-o', is_flag=True, default=False)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def import_workspace_cli(api_client, source_path, target_path, language, format, overwrite): # NOQA
    """
    Imports a file from local to the Databricks workspace.

    The format is by default SOURCE. Possible formats are SOURCE, HTML, JUPYTER, and DBC. Each
    format is documented at
    https://docs.databricks.com/api/latest/workspace.html#notebookexportformat.
    """
    WorkspaceApi(api_client).import_workspace(source_path, target_path, language, format, overwrite) # NOQA


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Exports a file from the Databricks workspace.')
@click.argument('source_path')
@click.argument('target_path')
@click.option('--format', '-f', default=WorkspaceFormat.SOURCE, type=FormatClickType())
@click.option('--overwrite', '-o', is_flag=True, default=False)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def export_workspace_cli(api_client, source_path, target_path, format, overwrite): # NOQA
    """
    Exports a notebook from the Databricks workspace.

    The format is by default SOURCE. Possible formats are SOURCE, HTML, JUPYTER, and DBC. Each
    format is documented at
    https://docs.databricks.com/api/latest/workspace.html#notebookexportformat.
    """
    if os.path.isdir(target_path):
        file_info = WorkspaceApi(api_client).get_status(source_path)
        if not file_info.is_notebook:
            raise RuntimeError('Export can only be called on a notebook.')
        extension = WorkspaceLanguage.to_extension(file_info.language)
        target_path = os.path.join(target_path, file_info.basename + extension)
    WorkspaceApi(api_client).export_workspace(source_path, target_path, format, overwrite) # NOQA


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Deletes objects from the Databricks workspace. '
                          'rm and delete are synonyms.')
@click.argument('workspace_path')
@click.option('--recursive', '-r', is_flag=True, default=False)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_cli(api_client, workspace_path, recursive):
    """
    Deletes objects from the Databricks workspace.

    To delete a folder add the recursive flag.
    """
    WorkspaceApi(api_client).delete(workspace_path, recursive)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Recursively exports a directory from the Databricks workspace.')
@click.argument('source_path')
@click.argument('target_path')
@click.option('--overwrite', '-o', is_flag=True, default=False)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def export_dir_cli(api_client, source_path, target_path, overwrite):
    """
    Recursively exports a directory from the Databricks workspace.

    Only directories and notebooks are exported. Notebooks are always exported in the SOURCE
    format. Notebooks will also have the extension of .scala, .py, .sql, or .r appended
    depending on the language type.
    """
    workspace_api = WorkspaceApi(api_client)
    assert workspace_api.get_status(source_path).is_dir, 'The source path must be a directory. {}' \
        .format(source_path)
    workspace_api.export_workspace_dir(source_path, target_path, overwrite)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help='Recursively imports a directory to the Databricks workspace.')
@click.argument('source_path')
@click.argument('target_path')
@click.option('--overwrite', '-o', is_flag=True, default=False)
@click.option('--exclude-hidden-files', '-e', is_flag=True, default=False)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def import_dir_cli(api_client, source_path, target_path, overwrite, exclude_hidden_files):
    """
    Recursively imports a directory from local to the Databricks workspace.

    Only directories and files with the extensions .scala, .py, .sql, .r, .R, .ipynb are imported.
    When imported, these extensions will be stripped off the name of the notebook.
    """
    WorkspaceApi(api_client).import_workspace_dir(source_path, target_path, overwrite,
                                                  exclude_hidden_files)


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help='Utility to interact with the Databricks workspace.')
@click.option('--version', '-v', is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
def workspace_group():  # pragma: no cover
    """
    Utility to interact with the Databricks workspace.
    Workspace paths must be absolute and be prefixed with `/`.
    """
    pass


workspace_group.add_command(ls_cli, name='ls')
workspace_group.add_command(ls_cli, name='list')
workspace_group.add_command(mkdirs_cli, name='mkdirs')
workspace_group.add_command(import_workspace_cli, name='import')
workspace_group.add_command(export_workspace_cli, name='export')
workspace_group.add_command(delete_cli, name='delete')
workspace_group.add_command(delete_cli, name='rm')
workspace_group.add_command(export_dir_cli, name='export_dir')
workspace_group.add_command(import_dir_cli, name='import_dir')



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base64 import b64encode, b64decode

import os
import shutil
import tempfile

import re
import click

from requests.exceptions import HTTPError

from databricks_cli.sdk import DbfsService
from databricks_cli.utils import error_and_quit
from databricks_cli.dbfs.dbfs_path import DbfsPath
from databricks_cli.dbfs.exceptions import LocalFileExistsException

BUFFER_SIZE_BYTES = 2**20


class ParseException(Exception):
    pass


class FileInfo(object):
    def __init__(self, dbfs_path, is_dir, file_size, modification_time):
        self.dbfs_path = dbfs_path
        self.is_dir = is_dir
        self.file_size = file_size
        self.modification_time = modification_time

    def to_row(self, is_long_form, is_absolute):
        path = self.dbfs_path.absolute_path if is_absolute else self.dbfs_path.basename
        stylized_path = click.style(path, 'cyan') if self.is_dir else path
        if is_long_form:
            filetype = 'dir' if self.is_dir else 'file'
            row = [filetype, self.file_size, stylized_path]
            # Add modification time if it is available.
            if self.modification_time is not None:
                row.append(self.modification_time)
            return row
        return [stylized_path]

    @classmethod
    def from_json(cls, json):
        dbfs_path = DbfsPath.from_api_path(json['path'])
        # If JSON doesn't include modification_time data, replace it with None.
        modification_time = json['modification_time'] if 'modification_time' in json else None
        return cls(dbfs_path, json['is_dir'], json['file_size'], modification_time)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.dbfs_path == other.dbfs_path and \
                self.is_dir == other.is_dir and \
                self.file_size == other.file_size and \
                self.modification_time == other.modification_time
        return False


class DbfsErrorCodes(object):
    RESOURCE_DOES_NOT_EXIST = 'RESOURCE_DOES_NOT_EXIST'
    RESOURCE_ALREADY_EXISTS = 'RESOURCE_ALREADY_EXISTS'
    PARTIAL_DELETE = 'PARTIAL_DELETE'


class DbfsApi(object):
    MULTIPART_UPLOAD_LIMIT = 2147483648

    def __init__(self, api_client):
        self.client = DbfsService(api_client)

    def list_files(self, dbfs_path, headers=None):
        list_response = self.client.list(dbfs_path.absolute_path, headers=headers)
        if 'files' in list_response:
            return [FileInfo.from_json(f) for f in list_response['files']]
        else:
            return []

    def file_exists(self, dbfs_path, headers=None):
        try:
            self.get_status(dbfs_path, headers=headers)
        except HTTPError as e:
            try:
                if e.response.json()['error_code'] == DbfsErrorCodes.RESOURCE_DOES_NOT_EXIST:
                    return False
            except ValueError:
                pass

            raise e
        return True

    def get_status(self, dbfs_path, headers=None):
        json = self.client.get_status(dbfs_path.absolute_path, headers=headers)
        return FileInfo.from_json(json)

    # Method makes multipart/form-data file upload for files <2GB.
    # Otherwise uses create, add-block, close methods for streaming upload.
    def put_file(self, src_path, dbfs_path, overwrite, headers=None):
        # If file size is >2Gb use streaming upload.
        if os.path.getsize(src_path) < self.MULTIPART_UPLOAD_LIMIT:
            self.client.put(dbfs_path.absolute_path, src_path=src_path,
                            overwrite=overwrite, headers=headers)
        else:
            handle = self.client.create(dbfs_path.absolute_path, overwrite,
                                        headers=headers)['handle']
            with open(src_path, 'rb') as local_file:
                while True:
                    contents = local_file.read(BUFFER_SIZE_BYTES)
                    if len(contents) == 0:
                        break
                    # add_block should not take a bytes object.
                    self.client.add_block(handle, b64encode(contents).decode(), headers=headers)
                self.client.close(handle, headers=headers)

    def get_file(self, dbfs_path, dst_path, overwrite, headers=None):
        if os.path.exists(dst_path) and not overwrite:
            raise LocalFileExistsException('{} exists already.'.format(dst_path))
        file_info = self.get_status(dbfs_path, headers=headers)
        if file_info.is_dir:
            error_and_quit(('The dbfs file {} is a directory.').format(repr(dbfs_path)))
        length = file_info.file_size
        offset = 0
        with open(dst_path, 'wb') as local_file:
            while offset < length:
                response = self.client.read(dbfs_path.absolute_path, offset, BUFFER_SIZE_BYTES,
                                            headers=headers)
                bytes_read = response['bytes_read']
                data = response['data']
                offset += bytes_read
                local_file.write(b64decode(data))

    @staticmethod
    def get_num_files_deleted(partial_delete_error):
        try:
            message = partial_delete_error.response.json()['message']
        except (AttributeError, KeyError):
            raise ParseException("Unable to retrieve the number of deleted files.")
        m = re.compile(r".*operation has deleted (\d+) files.*").match(message)
        if not m:
            raise ParseException(
                "Unable to retrieve the number of deleted files from the error message: {}".format(
                    message))
        return int(m.group(1))

    def delete(self, dbfs_path, recursive, headers=None):
        num_files_deleted = 0
        while True:
            try:
                self.client.delete(dbfs_path.absolute_path, recursive=recursive, headers=headers)
            except HTTPError as e:
                if e.response.status_code == 503:
                    try:
                        error_code = e.response.json()['error_code']
                    except (AttributeError, KeyError):
                        error_code = None
                    # Handle partial delete exceptions: retry until all the files have been deleted
                    if error_code == DbfsErrorCodes.PARTIAL_DELETE:
                        try:
                            num_files_deleted += DbfsApi.get_num_files_deleted(e)
                            click.echo("\rDeleted {} files. Delete in progress...\033[K".format(
                                num_files_deleted), nl=False)
                        except ParseException:
                            click.echo("\rDelete in progress...\033[K", nl=False)
                        continue
                click.echo("\rDeleted at least {} files but interrupted by error.\033[K".format(
                    num_files_deleted))
                raise e
            break
        click.echo("\rDelete finished successfully.\033[K")

    def mkdirs(self, dbfs_path, headers=None):
        self.client.mkdirs(dbfs_path.absolute_path, headers=headers)

    def move(self, dbfs_src, dbfs_dst, headers=None):
        self.client.move(dbfs_src.absolute_path, dbfs_dst.absolute_path, headers=headers)

    def _copy_to_dbfs_non_recursive(self, src, dbfs_path_dst, overwrite, headers=None):
        # Munge dst path in case dbfs_path_dst is a dir
        try:
            if self.get_status(dbfs_path_dst, headers=headers).is_dir:
                dbfs_path_dst = dbfs_path_dst.join(os.path.basename(src))
        except HTTPError as e:
            if e.response.json()['error_code'] == DbfsErrorCodes.RESOURCE_DOES_NOT_EXIST:
                pass
            else:
                raise e
        self.put_file(src, dbfs_path_dst, overwrite, headers=headers)

    def _copy_from_dbfs_non_recursive(self, dbfs_path_src, dst, overwrite, headers=None):
        # Munge dst path in case dst is a dir
        if os.path.isdir(dst):
            dst = os.path.join(dst, dbfs_path_src.basename)
        self.get_file(dbfs_path_src, dst, overwrite, headers=headers)

    def _copy_to_dbfs_recursive(self, src, dbfs_path_dst, overwrite, headers=None):
        try:
            self.mkdirs(dbfs_path_dst, headers=headers)
        except HTTPError as e:
            if e.response.json()['error_code'] == DbfsErrorCodes.RESOURCE_ALREADY_EXISTS:
                click.echo(e.response.json())
                return
        for filename in os.listdir(src):
            cur_src = os.path.join(src, filename)
            cur_dbfs_dst = dbfs_path_dst.join(filename)
            if os.path.isdir(cur_src):
                self._copy_to_dbfs_recursive(cur_src, cur_dbfs_dst, overwrite, headers=headers)
            elif os.path.isfile(cur_src):
                try:
                    self.put_file(cur_src, cur_dbfs_dst, overwrite, headers=headers)
                    click.echo('{} -> {}'.format(cur_src, cur_dbfs_dst))
                except HTTPError as e:
                    if e.response.json()['error_code'] == DbfsErrorCodes.RESOURCE_ALREADY_EXISTS:
                        click.echo('{} already exists. Skip.'.format(cur_dbfs_dst))
                    else:
                        raise e

    def _copy_from_dbfs_recursive(self, dbfs_path_src, dst, overwrite, headers=None):
        if os.path.isfile(dst):
            click.echo(
                '{} exists as a file. Skipping this subtree {}'.format(dst, repr(dbfs_path_src)))
            return
        elif not os.path.isdir(dst):
            os.makedirs(dst)

        for dbfs_src_file_info in self.list_files(dbfs_path_src, headers=headers):
            cur_dbfs_src = dbfs_src_file_info.dbfs_path
            cur_dst = os.path.join(dst, cur_dbfs_src.basename)
            if dbfs_src_file_info.is_dir:
                self._copy_from_dbfs_recursive(cur_dbfs_src, cur_dst, overwrite, headers=headers)
            else:
                try:
                    self.get_file(cur_dbfs_src, cur_dst, overwrite, headers=headers)
                    click.echo('{} -> {}'.format(cur_dbfs_src, cur_dst))
                except LocalFileExistsException:
                    click.echo(('{} already exists locally as {}. Skip. To overwrite, you ' +
                                'should provide the --overwrite flag.').format(cur_dbfs_src,
                                                                               cur_dst))

    def cp(self, recursive, overwrite, src, dst, headers=None):
        if not DbfsPath.is_valid(src) and DbfsPath.is_valid(dst):
            if not os.path.exists(src):
                error_and_quit('The local file {} does not exist.'.format(src))
            if not recursive:
                if os.path.isdir(src):
                    error_and_quit(
                        ('The local file {} is a directory. You must provide --recursive')
                        .format(src))
                self._copy_to_dbfs_non_recursive(src, DbfsPath(dst), overwrite, headers=headers)
            else:
                if not os.path.isdir(src):
                    self._copy_to_dbfs_non_recursive(src, DbfsPath(dst), overwrite, headers=headers)
                    return
                self._copy_to_dbfs_recursive(src, DbfsPath(dst), overwrite, headers=headers)
        # Copy from DBFS in this case
        elif DbfsPath.is_valid(src) and not DbfsPath.is_valid(dst):
            if not recursive:
                self._copy_from_dbfs_non_recursive(DbfsPath(src), dst, overwrite, headers=headers)
            else:
                dbfs_path_src = DbfsPath(src)
                if not self.get_status(dbfs_path_src, headers=headers).is_dir:
                    self._copy_from_dbfs_non_recursive(dbfs_path_src, dst, overwrite,
                                                       headers=headers)
                self._copy_from_dbfs_recursive(dbfs_path_src, dst, overwrite, headers=headers)
        elif not DbfsPath.is_valid(src) and not DbfsPath.is_valid(dst):
            error_and_quit('Both paths provided are from your local filesystem. '
                           'To use this utility, one of the src or dst must be prefixed '
                           'with dbfs:/')
        elif DbfsPath.is_valid(src) and DbfsPath.is_valid(dst):
            with TempDir() as temp_dir:
                # Always copy to <temp_dir>/temp since this will work no matter if it's a
                # recursive or a non-recursive copy.
                temp_path = temp_dir.path('temp')
                self.cp(recursive, True, src, temp_path)
                self.cp(recursive, overwrite, temp_path, dst)
        else:
            assert False, 'not reached'

    def cat(self, src):
        with TempDir() as temp_dir:
            temp_path = temp_dir.path('temp')
            self.cp(False, True, src, temp_path)
            with open(temp_path) as f:
                click.echo(f.read(), nl=False)


class TempDir(object):
    def __init__(self, remove_on_exit=True):
        self._dir = None
        self._path = None
        self._remove = remove_on_exit

    def __enter__(self):
        self._path = os.path.abspath(tempfile.mkdtemp())
        assert os.path.exists(self._path)
        return self

    def __exit__(self, tp, val, traceback):
        if self._remove and os.path.exists(self._path):
            shutil.rmtree(self._path)

        assert not self._remove or not os.path.exists(self._path)
        assert os.path.exists(os.getcwd())

    def path(self, *path):
        return os.path.join(self._path, *path)



"""Provide the API methods for GROUPs Databricks REST endpoint."""
# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from databricks_cli.click_types import OneOfOption
from databricks_cli.groups.api import GroupsApi
from databricks_cli.utils import eat_exceptions, CONTEXT_SETTINGS, pretty_format
from databricks_cli.configure.config import provide_api_client, profile_option, debug_option
from databricks_cli.version import print_version_callback, version


MEMBER_OPTIONS = ['user-name', 'group-name']


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help="Add an existing principal to another existing group.")
@click.option("--parent-name", required=True,
              help=("Name of the parent group to which the new member will be "
                    " added. This field is required."))
@click.option("--user-name", cls=OneOfOption, default=None, one_of=MEMBER_OPTIONS,
              help="The user name which will be added to the parent group.")
@click.option("--group-name", cls=OneOfOption, default=None, one_of=MEMBER_OPTIONS,
              help="If group name which will be added to the parent group.")
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def add_member_cli(api_client, parent_name, user_name, group_name):
    """Add a user or group to a group."""
    GroupsApi(api_client).add_member(parent_name=parent_name,
                                     user_name=user_name,
                                     group_name=group_name)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help="Create a new group with the given name.")
@click.option("--group-name", required=True)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def create_cli(api_client, group_name):
    """Create a new group with the given name."""
    content = GroupsApi(api_client).create(group_name)
    click.echo(pretty_format(content))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help="Return all of the members of a particular group.")
@click.option("--group-name", required=True)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_members_cli(api_client, group_name):
    """Return all of the members of a particular group."""
    content = GroupsApi(api_client).list_members(group_name)
    click.echo(pretty_format(content))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help="Return all of the groups in a workspace.")
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_all_cli(api_client):
    """Return all of the groups in an organization."""
    content = GroupsApi(api_client).list_all()
    click.echo(pretty_format(content))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help="Retrieve all groups in which a given user or group is a member.")
@click.option("--user-name", cls=OneOfOption, default=None, one_of=MEMBER_OPTIONS)
@click.option("--group-name", cls=OneOfOption, default=None, one_of=MEMBER_OPTIONS)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def list_parents_cli(api_client, user_name, group_name):
    """Retrieve all groups in which a given user or group is a member."""
    content = GroupsApi(api_client).list_parents(user_name=user_name,
                                                 group_name=group_name)
    click.echo(pretty_format(content))


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help="Removes a user or group from a group.")
@click.option("--parent-name", required=True,
              help=("Name of the parent group to which the new member will be "
                    " removed. This field is required."))
@click.option("--user-name", cls=OneOfOption, default=None, one_of=MEMBER_OPTIONS,
              help="The user name which will be removed from the parent group.")
@click.option("--group-name", cls=OneOfOption, default=None, one_of=MEMBER_OPTIONS,
              help="If group name which will be removed from the parent group.")
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def remove_member_cli(api_client, parent_name, user_name, group_name):
    GroupsApi(api_client).remove_member(parent_name=parent_name,
                                        user_name=user_name,
                                        group_name=group_name)


@click.command(context_settings=CONTEXT_SETTINGS,
               short_help="Remove a group from this organization.")
@click.option("--group-name", required=False)
@debug_option
@profile_option
@eat_exceptions
@provide_api_client
def delete_cli(api_client, group_name):
    """Remove a group from this organization."""
    content = GroupsApi(api_client).delete(group_name)
    click.echo(pretty_format(content))


@click.group(context_settings=CONTEXT_SETTINGS,
             short_help="Utility to interact with Databricks groups.")
@click.option("--version", "-v", is_flag=True, callback=print_version_callback,
              expose_value=False, is_eager=True, help=version)
@debug_option
@profile_option
@eat_exceptions
def groups_group():  # pragma: no cover
    """Provide utility to interact with Databricks groups."""
    pass


groups_group.add_command(add_member_cli, name="add-member")
groups_group.add_command(create_cli, name="create")
groups_group.add_command(list_members_cli, name="list-members")
groups_group.add_command(list_all_cli, name="list")
groups_group.add_command(list_parents_cli, name="list-parents")
groups_group.add_command(remove_member_cli, name="remove-member")
groups_group.add_command(delete_cli, name="delete")



# Databricks CLI
# Copyright 2017 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import click

from click import ParamType

from databricks_cli.utils import error_and_quit


class DbfsPath(object):
    def __init__(self, absolute_path, validate=True):
        self.absolute_path = absolute_path
        if validate:
            self.validate()

    @classmethod
    def from_api_path(cls, path):
        return cls('dbfs:' + path)

    @classmethod
    def is_valid(cls, path):
        return cls(path, validate=False).is_absolute_path

    def validate(self):
        """
        Checks that the path is a proper DbfsPath. it must have a prefix of
        "dbfs:" and must be an absolute path.
        """
        if self.absolute_path.startswith('dbfs://'):
            error_and_quit(('The path {} cannot start with dbfs://. '
                           'It must start with dbfs:/').format(repr(self)))
        if not self.is_absolute_path:
            error_and_quit('The path {} must start with "{}"'.format(
                repr(self), repr(DbfsPath('dbfs:/'))))

    def join(self, file_name):
        """
        Returns a new extended DBFS path with a file_name.
        :param: file_name
        :type: str
        :rtype: DbfsPath
        """
        stripped_dbfs_path = self._strip_trailing_slash()
        if stripped_dbfs_path.is_root:
            absolute_path = stripped_dbfs_path.absolute_path + file_name
        else:
            absolute_path = stripped_dbfs_path.absolute_path + '/' + file_name
        return DbfsPath(absolute_path)

    def relpath(self, dbfs_path_prefix):
        """
        Strips the prefix of this DbfsPath. Behaves very similar to os.path.relpath
        """
        return os.path.relpath(self.absolute_path, dbfs_path_prefix.absolute_path)

    @property
    def basename(self):
        """
        This function is like the basename utility and is unlike os.path.basename.
        >>> assert DbfsPath('dbfs:/').basename == ''
        >>> assert DbfsPath('dbfs:/test').basename == 'test'
        >>> assert DbfsPath('dbfs:/test/').basename == 'test'
        """
        if self.is_root:
            return ''
        elif self.absolute_path[-1] == '/':
            return self.absolute_path.split('/')[-2]
        else:
            return self.absolute_path.split('/')[-1]

    @property
    def is_absolute_path(self):
        return self.absolute_path.startswith('dbfs:/')

    @property
    def is_root(self):
        return self.absolute_path == 'dbfs:/'

    def _strip_trailing_slash(self):
        if self.is_root:
            return DbfsPath('dbfs:/')
        elif self.absolute_path[-1] == '/':
            return DbfsPath(self.absolute_path[0:-1])
        else:
            return DbfsPath(self.absolute_path)

    def __repr__(self):
        return click.style(self.absolute_path, underline=True)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.absolute_path == other.absolute_path
        return False


class DbfsPathClickType(ParamType):
    name = 'Path'

    def convert(self, value, param, ctx):
        return DbfsPath(value)



# Databricks CLI
# Copyright 2018 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
from datetime import datetime
from requests.exceptions import HTTPError

import click

from databricks_cli.jobs.api import JobsApi
from databricks_cli.workspace.api import WorkspaceApi, DIRECTORY, NOTEBOOK
from databricks_cli.dbfs.api import DbfsApi
from databricks_cli.workspace.types import WorkspaceLanguage
from databricks_cli.version import version as CLI_VERSION
from databricks_cli.stack.exceptions import StackError

MS_SEC = 1000

# Resource Services
JOBS_SERVICE = 'jobs'
WORKSPACE_SERVICE = 'workspace'
DBFS_SERVICE = 'dbfs'

# Config Outer Fields
STACK_NAME = 'name'
STACK_RESOURCES = 'resources'
STACK_DEPLOYED = 'deployed_resources'  # For Stack Status

# Resource Fields
RESOURCE_ID = 'id'
RESOURCE_SERVICE = 'service'
RESOURCE_WRITE_STATUS = 'writeStatus'
RESOURCE_PROPERTIES = 'properties'

# Resource Status Fields
RESOURCE_DATABRICKS_ID = 'databricks_id'
CLI_VERSION_KEY = 'cli_version'

# Job Service Properties
JOBS_RESOURCE_NAME = 'name'
JOBS_RESOURCE_JOB_ID = 'job_id'

# Workspace Service Properties
WORKSPACE_RESOURCE_SOURCE_PATH = 'source_path'
WORKSPACE_RESOURCE_PATH = 'path'
WORKSPACE_RESOURCE_OBJECT_TYPE = 'object_type'

# DBFS Service Properties
DBFS_RESOURCE_SOURCE_PATH = 'source_path'
DBFS_RESOURCE_PATH = 'path'
DBFS_RESOURCE_IS_DIR = 'is_dir'


class StackApi(object):
    def __init__(self, api_client):
        self.jobs_client = JobsApi(api_client)
        self.workspace_client = WorkspaceApi(api_client)
        self.dbfs_client = DbfsApi(api_client)

    def deploy(self, stack_config, stack_status=None, headers=None, **kwargs):
        """
        Deploys a stack given stack JSON configuration template at path config_path.

        After going through each of the resources and deploying them, stores status JSON
        of deployment with deploy status of each resource deployment.
        For each resource deployment, stack_status is used to get the associated resource status
        of a resource from the last deployment.

        :param stack_config: Must have the fields of
        'name', the name of the stack and 'resources', a list of stack resources.
        :param stack_status: Must have the fields of 'name', the name of the stack, 'resources',
        a list of stack resources, and 'deployed', a list of resource statuses from a previous
        deployment.
        :return: new_stack_status: The new stack status generated from the deployment of
        the given stack_config.
        """
        click.echo('#' * 80)
        self._validate_config(stack_config)
        if stack_status:
            click.echo('#' * 80)
            self._validate_status(stack_status)
            resource_id_to_status = self._get_resource_to_status_map(stack_status)
        else:
            resource_id_to_status = {}

        stack_name = stack_config.get(STACK_NAME)
        click.echo('#' * 80)
        click.echo('Deploying stack {}'.format(stack_name))

        # List of statuses, One for each resource in stack_config[STACK_RESOURCES]
        resource_statuses = []
        click.echo('#' * 80)
        for resource_config in stack_config.get(STACK_RESOURCES):
            # Retrieve resource deployment info from the last deployment.
            resource_map_key = (resource_config.get(RESOURCE_ID),
                                resource_config.get(RESOURCE_SERVICE))
            resource_status = resource_id_to_status.get(resource_map_key) \
                if resource_map_key in resource_id_to_status else None
            # Deploy resource, get resource_status
            new_resource_status = self._deploy_resource(resource_config,
                                                        resource_status,
                                                        headers=headers,
                                                        **kwargs)

            if resource_config.get(RESOURCE_WRITE_STATUS, True):
                resource_statuses.append(new_resource_status)
            click.echo('#' * 80)

        new_stack_status = {STACK_NAME: stack_name,
                            CLI_VERSION_KEY: CLI_VERSION,
                            STACK_DEPLOYED: resource_statuses}

        # Validate that the status has been created correctly
        self._validate_status(new_stack_status)
        click.echo('#' * 80)

        return new_stack_status

    def download(self, stack_config, headers=None, **kwargs):
        """
        Downloads a stack given a dict of the stack configuration.
        :param stack_config: dict of stack configuration. Must contain 'name' and 'resources' field.
        :return: None.
        """
        self._validate_config(stack_config)
        stack_name = stack_config.get(STACK_NAME)
        click.echo('Downloading stack {}'.format(stack_name))

        click.echo('#' * 80)
        for resource_config in stack_config.get(STACK_RESOURCES):
            # Deploy resource, get resource_status
            self._download_resource(resource_config, headers=headers, **kwargs)
            click.echo('#' * 80)

    def _deploy_resource(self, resource_config, resource_status=None, headers=None, **kwargs):
        """
        Deploys a resource given a resource information extracted from the stack JSON configuration
        template.

        :param resource_config: A dict of the resource with fields of 'id', 'service'
        and 'properties'.
        ex. {'id': 'example-resource', 'service': 'jobs', 'properties': {...}}
        :param resource_status: A dict of the resource's deployment info from the last
        deployment. Will be None if this is the first deployment.
        ex. {'id': 'example-resource', 'service': 'jobs', 'databricks_id': {...}}
        :return: dict resource_status- A dictionary of deployment information of the
        resource to be stored at deploy time. It includes the resource id of the resource along
        with the databricks id and deploy output of the resource.
        ex. {'id': 'example-resource', 'service': 'jobs', 'databricks_id': {'job_id': 123}}
        """
        resource_id = resource_config.get(RESOURCE_ID)
        resource_service = resource_config.get(RESOURCE_SERVICE)
        resource_properties = resource_config.get(RESOURCE_PROPERTIES)
        databricks_id = resource_status.get(RESOURCE_DATABRICKS_ID) if resource_status else None

        if resource_service == JOBS_SERVICE:
            click.echo('Deploying job "{}" with properties: \n{}'.format(resource_id, json.dumps(
                resource_properties, indent=2, separators=(',', ': '))))
            new_databricks_id = self._deploy_job(resource_properties,
                                                 databricks_id,
                                                 headers=headers)
        elif resource_service == WORKSPACE_SERVICE:
            click.echo(
                'Deploying workspace asset "{}" with properties \n{}'
                .format(
                    resource_id, json.dumps(resource_properties, indent=2, separators=(',', ': '))
                )
            )
            overwrite = kwargs.get('overwrite', False)
            new_databricks_id = self._deploy_workspace(resource_properties,
                                                       databricks_id,
                                                       overwrite,
                                                       headers=headers)
        elif resource_service == DBFS_SERVICE:
            click.echo(
                'Deploying DBFS asset "{}" with properties \n{}'.format(
                    resource_id, json.dumps(resource_properties, indent=2, separators=(',', ': '))
                )
            )
            overwrite = kwargs.get('overwrite', False)
            new_databricks_id = self._deploy_dbfs(resource_properties,
                                                  databricks_id,
                                                  overwrite,
                                                  headers=headers)
        else:
            raise StackError('Resource service "{}" not supported'.format(resource_service))

        new_resource_status = {RESOURCE_ID: resource_id,
                               RESOURCE_SERVICE: resource_service,
                               RESOURCE_DATABRICKS_ID: new_databricks_id}
        return new_resource_status

    def _download_resource(self, resource_config, headers=None, **kwargs):
        """
        Downloads a resource given a resource information extracted from the stack JSON
        configuration template.

        :param resource_config: A dict of the resource with fields of 'id', 'service' and
        'properties'.
        ex. {'id': 'example-resource', 'service': 'jobs', 'properties': {...}}
        """
        resource_id = resource_config.get(RESOURCE_ID)
        resource_service = resource_config.get(RESOURCE_SERVICE)
        resource_properties = resource_config.get(RESOURCE_PROPERTIES)

        if resource_service == WORKSPACE_SERVICE:
            click.echo(
                'Downloading workspace asset "{}" with properties \n{}'
                .format(
                    resource_id, json.dumps(resource_properties, indent=2, separators=(',', ': '))
                )
            )
            overwrite = kwargs.get('overwrite', False)
            self._download_workspace(resource_properties, overwrite, headers=headers)
        else:
            click.echo('Resource service "{}" not supported for download. '
                       'skipping.'.format(resource_service))

    def _deploy_job(self, resource_properties, databricks_id=None, headers=None):
        """
        Deploys a job resource by either creating a job if the job isn't kept track of through
        the databricks_id of the job or updating an existing job. The job is created or updated
        using the the settings specified in the inputted job_settings.

        :param resource_properties: A dict of the Databricks JobSettings data structure
        :param databricks_id: A dict object containing 'job_id' field of job identifier in
        Databricks server

        :return: databricks_id: dict containing  a 'job_id' field of the physical job_id of the
        job on databricks.
        """
        job_settings = resource_properties  # resource_properties of jobs are solely job settings.

        if databricks_id:
            job_id = databricks_id.get(JOBS_RESOURCE_JOB_ID)
            self._update_job(job_settings, job_id, headers=headers)
        else:
            job_id = self._put_job(job_settings, headers=headers)
        click.echo("Job deployed on Databricks with Job ID {}".format(job_id))
        databricks_id = {JOBS_RESOURCE_JOB_ID: job_id}
        return databricks_id

    def _put_job(self, job_settings, headers=None):
        """
        Given settings of the job in job_settings, create a new job. For purposes of idempotency
        and to reduce leaked resources in alpha versions of stack deployment, if a job exists
        with the same name, that job will be updated. If multiple jobs are found with the same name,
        the deployment will abort.

        :param job_settings:
        :return: job_id, Physical ID of job on Databricks server.
        """
        job_name = job_settings.get(JOBS_RESOURCE_NAME)
        jobs_same_name = self.jobs_client._list_jobs_by_name(job_name, headers=headers)
        if len(jobs_same_name) > 1:
            raise StackError('Multiple jobs with the same name "{}" already exist, aborting'
                             ' stack deployment'.format(job_name))
        if len(jobs_same_name) == 1:
            existing_job = jobs_same_name[0]
            creator_name = existing_job.get('creator_user_name')
            timestamp = existing_job.get('created_time') / MS_SEC  # Convert to readable date.
            date_created = datetime.fromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')
            click.echo('Warning: Job exists with same name "{}" created by {} on {}. Job will '
                       'be overwritten'.format(job_name, creator_name, date_created))
            # Calling jobs_client.reset_job directly so as to not call same level function.
            self.jobs_client.reset_job({'job_id': existing_job.get('job_id'),
                                        'new_settings': job_settings}, headers=headers)
            return existing_job.get('job_id')
        else:
            job_id = self.jobs_client.create_job(job_settings, headers=headers).get('job_id')
            return job_id

    def _update_job(self, job_settings, job_id, headers=None):
        """
        Given job settings and an existing job_id of a job, update the job settings on databricks.

        :param job_settings: job settings to update the job with.
        :param job_id: physical job_id of job in databricks server.
        """
        try:
            self.jobs_client.reset_job({'job_id': job_id, 'new_settings': job_settings},
                                       headers=headers)
        except HTTPError:
            raise StackError('Job ID {} in stack status could not be found in the workspace. '
                             'Please remove or make necessary changes to the current stack status '
                             'to resolve this inconsistency before proceeding. Aborting '
                             'stack deployment ...'.format(job_id))

    def _deploy_workspace(self, resource_properties, databricks_id, overwrite, headers=None):
        """
        Deploy workspace asset.

        :param resource_properties: dict of properties for the workspace asset. Must contain the
        'source_path', 'path' and 'object_type' fields.
        :param databricks_id: dict containing physical identifier of workspace asset on databricks.
        Should contain the field 'path'.
        :param overwrite: Whether or not to overwrite the contents of workspace notebooks.
        :return:  databricks_id: dict containing the physical ID
        for the stack status that contains the workspace path of the notebook or directory on
        datbricks. deploy_output is the initial information about the asset on databricks at deploy
        time returned by the REST API.
        """
        local_path = resource_properties.get(WORKSPACE_RESOURCE_SOURCE_PATH)
        workspace_path = resource_properties.get(WORKSPACE_RESOURCE_PATH)
        object_type = resource_properties.get(WORKSPACE_RESOURCE_OBJECT_TYPE)

        actual_object_type = DIRECTORY if os.path.isdir(local_path) else NOTEBOOK
        if object_type != actual_object_type:
            raise StackError('Field "{}" ({}) not consistent '
                             'with actual object type ({})'.format(WORKSPACE_RESOURCE_OBJECT_TYPE,
                                                                   object_type,
                                                                   actual_object_type))

        click.echo('Uploading {} from {} to Databricks workspace at {}'.format(object_type,
                                                                               local_path,
                                                                               workspace_path))
        if object_type == NOTEBOOK:
            # Inference of notebook language and format
            language_fmt = WorkspaceLanguage.to_language_and_format(local_path)
            if language_fmt is None:
                raise StackError("Workspace notebook language and format cannot be inferred. "
                                 "Please check file extension of notebook file.")
            language, fmt = language_fmt
            # Create needed directories in workspace.
            self.workspace_client.mkdirs(os.path.dirname(workspace_path), headers=headers)
            self.workspace_client.import_workspace(local_path, workspace_path, language, fmt,
                                                   overwrite, headers=headers)
        elif object_type == DIRECTORY:
            self.workspace_client.import_workspace_dir(local_path, workspace_path, overwrite,
                                                       exclude_hidden_files=True, headers=headers)
        else:
            # Shouldn't reach here because of verification of object_type above.
            assert False

        if databricks_id and databricks_id[WORKSPACE_RESOURCE_PATH] != workspace_path:
            # databricks_id['path'] is the workspace path from the last deployment. Alert when
            # changed
            click.echo("Workspace asset had path changed from {} to {}"
                       .format(databricks_id[WORKSPACE_RESOURCE_PATH], workspace_path))
        new_databricks_id = {WORKSPACE_RESOURCE_PATH: workspace_path}

        return new_databricks_id

    def _download_workspace(self, resource_properties, overwrite, headers=None):
        """
        Download workspace asset.

        :param resource_properties: dict of properties for the workspace asset. Must contain the
        'source_path', 'path' and 'object_type' fields.
        :param overwrite: Whether or not to overwrite the contents of workspace notebooks.
        """
        local_path = resource_properties.get(WORKSPACE_RESOURCE_SOURCE_PATH)
        workspace_path = resource_properties.get(WORKSPACE_RESOURCE_PATH)
        object_type = resource_properties.get(WORKSPACE_RESOURCE_OBJECT_TYPE)
        click.echo('Downloading {} from Databricks path {} to {}'.format(object_type,
                                                                         workspace_path,
                                                                         local_path))
        if object_type == NOTEBOOK:
            # Inference of notebook language and format. A tuple of (language, fmt) or Nonetype.
            language_fmt = WorkspaceLanguage.to_language_and_format(local_path)
            if language_fmt is None:
                raise StackError("Workspace Notebook language and format cannot be inferred. "
                                 "Please check file extension of notebook 'source_path'.")
            (_, fmt) = language_fmt
            local_dir = os.path.dirname(os.path.abspath(local_path))
            if not os.path.exists(local_dir):
                os.makedirs(local_dir)
            self.workspace_client.export_workspace(workspace_path, local_path, fmt, overwrite,
                                                   headers=headers)
        elif object_type == DIRECTORY:
            self.workspace_client.export_workspace_dir(workspace_path, local_path, overwrite,
                                                       headers=headers)
        else:
            raise StackError('Invalid value for "{}" field: {}'
                             .format(WORKSPACE_RESOURCE_OBJECT_TYPE, object_type))

    def _deploy_dbfs(self, resource_properties, databricks_id, overwrite, headers=None):
        """
        Deploy dbfs asset.

        :param resource_properties: dict of properties for the dbfs asset. Must contain the
        'source_path', 'path' and 'is_dir' fields.
        :param databricks_id: dict containing physical identifier of dbfs asset on Databricks.
        Should contain the field 'path'.
        :param overwrite: Whether or not to overwrite the contents of dbfs files.
        :return: databricks_id:  a dict that contains the dbfs path of the file on Databricks.
        ex.{"path":"dbfs:/path/in/dbfs"}
        """

        local_path = resource_properties.get(DBFS_RESOURCE_SOURCE_PATH)
        dbfs_path = resource_properties.get(DBFS_RESOURCE_PATH)
        is_dir = resource_properties.get(DBFS_RESOURCE_IS_DIR)

        if is_dir != os.path.isdir(local_path):
            dir_or_file = 'directory' if os.path.isdir(local_path) else 'file'
            raise StackError('local source_path "{}" is found to be a {}, but is not specified'
                             ' as one with is_dir: {}.'
                             .format(local_path, dir_or_file, str(is_dir).lower()))
        if is_dir:
            click.echo('Uploading directory from {} to DBFS at {}'.format(local_path, dbfs_path))
            self.dbfs_client.cp(recursive=True, overwrite=overwrite, src=local_path, dst=dbfs_path,
                                headers=headers)
        else:
            click.echo('Uploading file from {} to DBFS at {}'.format(local_path, dbfs_path))
            self.dbfs_client.cp(recursive=False, overwrite=overwrite, src=local_path, dst=dbfs_path,
                                headers=headers)

        if databricks_id and databricks_id[DBFS_RESOURCE_PATH] != dbfs_path:
            # databricks_id['path'] is the dbfs path from the last deployment. Alert when changed
            click.echo("Dbfs asset had path changed from {} to {}"
                       .format(databricks_id[DBFS_RESOURCE_PATH], dbfs_path))
        new_databricks_id = {DBFS_RESOURCE_PATH: dbfs_path}

        return new_databricks_id

    def _validate_config(self, stack_config):
        """
        Validate fields within a stack configuration. This ensures that an inputted configuration
        has the necessary fields for stack deployment to function well.

        :param stack_config: dict- stack config that is inputted by the user.
        :return: None. Raises errors to stop deployment if there is a problem.
        """
        click.echo('Validating fields in stack configuration...')
        self._assert_fields_in_dict([STACK_NAME, STACK_RESOURCES], stack_config)

        seen_resource_ids = set()  # Store seen resources to restrict duplicates.
        for resource in stack_config.get(STACK_RESOURCES):
            # Get validate resource ID exists, then get it.
            self._assert_fields_in_dict([RESOURCE_ID], resource)
            resource_id = resource.get(RESOURCE_ID)

            click.echo('Validating fields in resource with ID "{}"'.format(resource_id))
            self._assert_fields_in_dict([RESOURCE_SERVICE, RESOURCE_PROPERTIES], resource)

            resource_service = resource.get(RESOURCE_SERVICE)
            resource_properties = resource.get(RESOURCE_PROPERTIES)

            # Error on duplicate resource ID's
            if resource_id in seen_resource_ids:
                raise StackError('Duplicate resource ID "{}" found, please resolve.'.format(
                    resource_id))
            seen_resource_ids.add(resource_id)

            # Resource service-specific validations
            click.echo('Validating fields in "{}" of {} resource.'
                       .format(RESOURCE_PROPERTIES, resource_service))
            if resource_service == JOBS_SERVICE:
                self._assert_fields_in_dict([JOBS_RESOURCE_NAME], resource_properties)
            elif resource_service == WORKSPACE_SERVICE:
                self._assert_fields_in_dict(
                    [WORKSPACE_RESOURCE_PATH, WORKSPACE_RESOURCE_SOURCE_PATH,
                     WORKSPACE_RESOURCE_OBJECT_TYPE], resource_properties)
            elif resource_service == DBFS_SERVICE:
                self._assert_fields_in_dict(
                    [DBFS_RESOURCE_PATH, DBFS_RESOURCE_SOURCE_PATH,
                     DBFS_RESOURCE_IS_DIR], resource_properties)
            else:
                raise StackError('Resource service "{}" not supported'.format(resource_service))

    def _validate_status(self, stack_status):
        """
        Validate fields within a stack status. This ensures that a stack status has the
        necessary fields for stack deployment to function well.

        If there is an error here, then it is either an implementation error that must be fixed by
        a developer or the User edited the stack status file created by the program.

        :param stack_status: dict- stack status that is created by the program.
        :return: None. Raises errors to stop deployment if there is a problem.
        """
        click.echo('Validating fields in stack status...')
        self._assert_fields_in_dict([STACK_NAME, STACK_DEPLOYED], stack_status)

        for resource_status in stack_status.get(STACK_DEPLOYED):
            self._assert_fields_in_dict([RESOURCE_ID], resource_status)
            resource_id = resource_status.get(RESOURCE_ID)
            click.echo('Validating fields in resource status of resource with ID "{}"'
                       .format(resource_id))
            self._assert_fields_in_dict([RESOURCE_SERVICE, RESOURCE_DATABRICKS_ID],
                                        resource_status)

            resource_service = resource_status.get(RESOURCE_SERVICE)
            resource_databricks_id = resource_status.get(RESOURCE_DATABRICKS_ID)

            click.echo('Validating fields in "{}" of {} resource status'
                       .format(RESOURCE_DATABRICKS_ID, resource_service))
            if resource_service == JOBS_SERVICE:
                self._assert_fields_in_dict([JOBS_RESOURCE_JOB_ID], resource_databricks_id)
            elif resource_service == WORKSPACE_SERVICE:
                self._assert_fields_in_dict([WORKSPACE_RESOURCE_PATH], resource_databricks_id)
            elif resource_service == DBFS_SERVICE:
                self._assert_fields_in_dict([DBFS_RESOURCE_PATH], resource_databricks_id)
            else:
                raise StackError("{} not a valid resource status service".format(resource_service))

    def _assert_fields_in_dict(self, fields, dictionary):
        for field in fields:
            if field not in dictionary:
                raise StackError('Required field "{}" not found'.format(field))

    def _get_resource_to_status_map(self, stack_status):
        """
        Returns a dictionary that maps a resource's (id, service) to the resource's status
        from the last deployment

        The key for this dictionary is the resource's (id, service) so that we don't load
        persisted resources with the wrong resource service.
        """
        return {
            (resource_status.get(RESOURCE_ID), resource_status.get(RESOURCE_SERVICE)):
                resource_status
            for resource_status in stack_status.get(STACK_DEPLOYED)
        }



# Databricks CLI
# Copyright 2018 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"), except
# that the use of services to which certain application programming
# interfaces (each, an "API") connect requires that the user first obtain
# a license for the use of the APIs from Databricks, Inc. ("Databricks"),
# by creating an account at www.databricks.com and agreeing to either (a)
# the Community Edition Terms of Service, (b) the Databricks Terms of
# Service, or (c) another written agreement between Licensee and Databricks
# for the use of the APIs.
#
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class StackError(Exception):
    pass
